# CLEANUP NOTES:
# - remove unused comments
# - add documentation for each method
# - add comments inline explaining each piece
# - add a unit test for each method (at least)

# stdlib
from functools import lru_cache
from typing import Any
from typing import Callable
from typing import Dict as TypeDict
from typing import List as TypeList
from typing import Optional
from typing import Set
from typing import Tuple as TypeTuple
from typing import Union

# third party
import numpy as np
from pymbolic.mapper import WalkMapper
from pymbolic.mapper.evaluator import EvaluationMapper as EM
from pymbolic.primitives import Variable
from scipy import optimize
import sympy as sym
from sympy.core.basic import Basic
from sympy.solvers import solve

# relative
from .entity import Entity

# Leaving this commented out here because I'm pretty sure we can get the
# lru_cache to be WAY faster through this approach but I can't seem to
# get it to work (it adds about 10% perf loss).
# ordered_symbols = list()
# for i in range(100):
#     ordered_symbols.append(symbols("s"+str(i)))

# ssid2obj is used to lookup value, min_val and max_val. if we are able to store these
# in the name string instead we could extract them at the point of use instead of lookup
# TypeDict[str, Union[PhiScalar, GammaScalar]]
ssid2obj: TypeDict[str, Any] = {}  # TODO: Fix types in circular deps


class GetSymbolsMapper(WalkMapper):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.free_symbols: Set[Any] = set()

    def visit(self, *args: Any, **kwargs: Any) -> bool:
        for arg in args:
            if isinstance(arg, Variable):
                self.free_symbols.add(arg)
        return True


def create_searchable_function_from_polynomial(
    poly: Basic, symbol2index: dict
) -> Callable:
    """Wrap polynomial execution logic in a function usable for scipy.optimize

    scipy.optimize functions expect an ordered list of args as input whereas
    sympy passes dictionaries into a .subs() function. This method just
    wraps sympy's approach in a method which accepts a tuple of args ordered
    according to symbol2index lookup table
    """
    if "pymbolic" in str(type(poly)):

        def _run_specific_args(tuple_of_args: TypeTuple) -> EM:
            kwargs = {sym: tuple_of_args[i] for sym, i in symbol2index.items()}
            output = EM(context=kwargs)(poly)
            return output

    else:

        def _run_specific_args(
            tuple_of_args: TypeTuple,
        ) -> EM:
            kwargs = {sym: tuple_of_args[i] for sym, i in symbol2index.items()}
            output = poly.subs(kwargs)

            return output

    return _run_specific_args


@lru_cache(maxsize=None)
def minimize_poly(
    poly: Basic, *rranges: TypeTuple, force_all_searches: bool = False, **s2i: TypeDict
) -> TypeList[optimize.OptimizeResult]:
    """Minimizes a polynomial using types basic enough for lru_cache"""

    # convert polynomial to function object with API necessary for scipy optimizer
    search_fun = create_searchable_function_from_polynomial(poly, s2i)

    # search over rranges for the set of inputs which minimizes function "f"
    return minimize_function(f=search_fun, rranges=rranges, force_all_searches=False)


def flatten_and_maximize_sympoly(
    poly: Basic, force_all_searches: bool = False
) -> TypeList[optimize.OptimizeResult]:

    i2s = list(poly.free_symbols)
    s2i = {s: i for i, s in enumerate(i2s)}

    # this code seems to make things slower - although there might be a memory
    # improvement (i haven't checked)
    # flattened_poly = poly.copy().subs({k:v for k,v in zip(i2s, ordered_symbols[0:len(i2s)])})
    # flattened_s2i = {str(ordered_symbols[i]):i for s,i in s2i.items()}

    flattened_poly = poly
    flattened_s2i = {str(s): i for s, i in s2i.items()}

    rranges = [
        (ssid2obj[i2s[i].name].min_val, ssid2obj[i2s[i].name].max_val)
        for i in range(len(s2i))
    ]

    return minimize_poly(
        flattened_poly, *rranges, force_all_searches=force_all_searches, **flattened_s2i
    )


def flatten_and_maximize_poly(
    poly: Any, force_all_searches: bool = False
) -> TypeList[optimize.OptimizeResult]:

    mapper = GetSymbolsMapper()
    mapper(poly)

    i2s = list(mapper.free_symbols)
    s2i = {s: i for i, s in enumerate(i2s)}

    # this code seems to make things slower - although there might be a memory
    # improvement (i haven't checked)
    # flattened_poly = poly.copy().subs({k:v for k,v in zip(i2s, ordered_symbols[0:len(i2s)])})
    # flattened_s2i = {str(ordered_symbols[i]):i for s,i in s2i.items()}

    flattened_poly = poly
    flattened_s2i = {str(s): i for s, i in s2i.items()}

    rranges = [
        (ssid2obj[i2s[i].name].min_val, ssid2obj[i2s[i].name].max_val)
        for i in range(len(s2i))
    ]

    return minimize_poly(
        flattened_poly, *rranges, force_all_searches=force_all_searches, **flattened_s2i
    )


def create_lookup_tables_for_symbol(
    polynomial: Any,
) -> TypeTuple[TypeList[str], TypeDict[str, int]]:

    mapper = GetSymbolsMapper()
    mapper(polynomial)

    index2symbol = [str(x) for x in mapper.free_symbols]
    symbol2index = {sym: i for i, sym in enumerate(index2symbol)}

    return index2symbol, symbol2index


def minimize_function(
    f: Any,
    rranges: Any,
    constraints: Optional[TypeList[TypeDict[str, Any]]] = None,
    force_all_searches: bool = False,
) -> TypeList[optimize.OptimizeResult]:
    constraints = constraints if constraints is not None else []

    results = list()

    # Step 1: try simplicial
    shgo_results = optimize.shgo(
        f, rranges, sampling_method="simplicial", constraints=constraints
    )
    results.append(shgo_results)

    if not shgo_results.success or force_all_searches:
        # sometimes simplicial has trouble as a result of initialization
        # see: https://github.com/scipy/scipy/issues/10429 for details
        shgo_results = optimize.shgo(
            f, rranges, sampling_method="sobol", constraints=constraints
        )
        results.append(shgo_results)

    if not shgo_results.success:
        raise Exception("Search algorithm wasn't solvable... abort")

    return results


def max_lipschitz_wrt_entity(scalars: Any, entity: Entity) -> float:
    # if all scalars have is_linear = True, we will skip the search below
    can_skip = True
    for i in scalars:
        if getattr(i, "is_linear", None) is not True:
            can_skip = False
            break
    if can_skip:
        return 1.0

    result: Union[float, optimize.OptimizeResult] = max_lipschitz_via_jacobian(
        scalars, input_entity=entity
    )[0][-1]
    if not isinstance(result, float):
        return -float(result.fun)
    else:
        return -result


def max_lipschitz_via_jacobian(
    scalars: TypeList[Any],  # TODO: Fix Scalar type circular
    input_entity: Optional[Entity] = None,
    data_dependent: bool = True,
    force_all_searches: bool = False,
    try_hessian_shortcut: bool = False,
) -> TypeTuple[TypeList[float], Any]:
    # scalars = R^d` representing the d' dimensional output of g
    # input_entity = the 'i'th entity for which we want to compute a lipschitz bound

    input_scalars = set()
    for s in scalars:
        for i_s in s.input_scalars:
            input_scalars.add(i_s)

    # R^d` representing the d' dimensional output of g
    # the numberator of the partial derivative
    out = sym.Matrix([x.sympoly for x in scalars])

    if input_entity is None:
        # X_1 through X_n flattened into a single vector
        j = out.jacobian([x.sympoly for x in input_scalars])
    else:

        # In general it doesn't make sense to consider the max partial derivative over
        # all inputs because we don't want the Lipschitz bound of the entire jacobian,
        # we want the Lipschitz bound with respect to entity "i" (see Individual Privacy
        # Accounting via a Renyi Filter: https://arxiv.org/abs/2008.11193). For example,
        # if I had a polynomial y = a + b**2 + c**3 + d**4 where each a,b,c,d variable
        # was from a different entity, the fact that d has a big derivative should
        # change the max Lipschitz bound of y with respect to "a". Thus, we're only
        # interested in searching for the maximum partial derivative with respect to the
        # variables from the focus entity "i".

        # And if we're looking to compute the max parital derivative with respect to
        # input scalars from only one entity, then we select only the variables
        # corresponding to that entity here.

        # X_i scalars
        relevant_scalars = list(
            filter(lambda s: s.entity == input_entity, input_scalars)
        )
        relevant_inputs = [x.sympoly for x in relevant_scalars]

        # jacobian ONLY with respect to X_i
        j = out.jacobian(relevant_inputs)

        # For higher order functions - it's possible that some of the partial
        # derivatives are conditioned on data from the input entity. The philosophy of
        # input DP is that when producing an epsilon guarantee for entity[i] that we
        # don't need to search over the possible range of data for that entity but can
        # instead use the data itself - this results in an epsilon for each entity which
        # is private but it also means the bound is tighter. So we could skip this step
        # but it would in some cases make the bound looser than it needs to be.
        if data_dependent:
            j = j.subs({x.sympoly: x.value for x in relevant_scalars})

    # WARNING: in the Feldman paper (https://arxiv.org/pdf/2008.11193.pdf) in
    # Example 2.8, which is the basis for this logic, the paper says the following,
    # "Suppose that g : (Rd)n → Rd` is Li-Lipschitz in coordinate i (in L2-norm)."
    # This is a somewhat ambiguous statement. After much thought, we are confident
    # that 'i' refers to the index of a specific entity, but it's unclear exactly
    # what is being L2-normed. There are two plausible options.

    # - Option A: each output value is L2 normed before the derivative is taken, meaning
    # that Li-Lipschitz is a bound the absolute value of the derivative of g wrt each input
    # from entity i. Note that this means that we took a vector of polynomials, L2normed htem,
    # took the derivative, and then the Li-lipschiktz bound is wrt each of these derivatives.
    # This seems a bit odd... because why L2 norm something when Lipschitz is going to calculate
    # the absolute value anyway (it's an identical operation... although absolute value isn't
    # technically continuous and differentiable in all places, so it could be that this is
    # express as L2 norm because that's the mathematically elegant way to do it. Also it makes more
    # sense why it would be a paranthetical given that it's just a small explanation as to why
    # the derivaitve of the L-Lipschitz bound can be taken). Perhaps this paraenthetical was
    # just a throwaway.

    # - Option B: more likely, because the Li-lipschitz constraint is referring to a vector
    # of derivatives corresponding to the input->output relationship between a single output
    # scalar and all of the inputs from that entity, the L2 norm is being used to compress
    # all of these derivatives into one scalar, which is bounded. That is to say, Option A
    # is bound on the derivative of the L2 norm of each value, whereas Option B is a bound
    # on the L2 norm of the derivatives, which is a scalar. We think that it's the latter
    # because it makes sense why the L2 norm is being applied (not redundant). The bigger reason
    # why Option B makes more sense is that with option A, if we publish an out value with
    # multiple inputs coming from one entity, the privacy budget is only scaled WRT the sharpest
    # derivative of one input as opposed to the combined effect of all of them. This follows
    # the differential privacy intuition that we care about the overall effect of removing
    # all inputs toa  function from entity, not just the max effect of any one of htose inputs.
    # Thus, Option B *should* be the right answer, and if it's not, then we probably end
    # up writing a paper for why Vitaly Feldman's paper is broken.

    # ON THE CONTRARY: in Example 2.8, the L term is just supposed to be a worst case scalar
    # on the function, which gets multipled by a (separate) L2 norm of the underlying data
    # values, so it makes sense that we wouldn't scale by the combined factor of all inputs, but
    # instead scale by the "worst case relationship" between any input and any output.

    # ON THE CONTRARY (Again): L_i is supposed to be the maximum slope WRT all contributions
    # from this entity, meaning the worst case bound of removing all inputs to this function.
    # thus it follows that the combined inputs could be worse than any particular input?...

    # CONSERVATIVE VIEW (and interim conclusion): Option B will always be higher than option A,
    # because after ensuring
    # that everything is positive value (by squaring it), Option A calculates an max across values
    # whereas option B sums them. This means that Option B will always generate at least the same
    # if not higher value than Option A, so given the uncertainty we're going with Option B.
    neg_l2_j = -((np.sum(np.square(j))) ** 0.5)

    # if the L2 norm is a constant (flat derivative) - just return it
    if len(np.sum(j).free_symbols) == 0:
        result = -float(np.max(j))
        return [result], neg_l2_j

    if try_hessian_shortcut:
        h = j.jacobian([x.sympoly for x in input_scalars])
        if (
            len(solve(np.sum(h**2), *[x.sympoly for x in input_scalars], dict=True))
            == 0
        ):
            print(
                "Gradient is linear - solve with brute force search over edges of domain"
            )

            i2s, s2i = create_lookup_tables_for_symbol(neg_l2_j)
            search_fun = create_searchable_function_from_polynomial(
                poly=neg_l2_j, symbol2index=s2i
            )

            constant = 0.000001
            rranges = [
                (x.min_val, x.max_val, x.max_val - x.min_val) for x in input_scalars
            ]
            skewed_results = optimize.brute(
                search_fun, rranges, finish=None, full_output=False
            )
            result_inputs = skewed_results + constant
            result_output = search_fun(result_inputs)
            return [float(result_output)], neg_l2_j

    return flatten_and_maximize_sympoly(neg_l2_j), neg_l2_j



# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

# relative
from .primitive_factory import PrimitiveFactory
from .primitive_factory import isprimitive


def downcast(value: Any, recurse: bool = True) -> Any:
    if isprimitive(value=value):
        # Wrap in a SyPrimitive
        return PrimitiveFactory.generate_primitive(value=value, recurse=recurse)
    else:
        return value


def downcast_args_and_kwargs(
    args: Union[List[Any], Tuple[Any, ...]], kwargs: Dict[Any, Any]
) -> Tuple[List[Any], Dict[Any, Any]]:
    # when we try to handle primitives they often need to be converted to PyPrimitive
    # first so they can be serialized and sent around
    downcasted_args = []
    downcasted_kwargs = {}
    for arg in args:
        # check if its primitive
        downcasted_args.append(downcast(value=arg))

    for k, arg in kwargs.items():
        # check if its primitive
        downcasted_kwargs[k] = downcast(value=arg)

    return (downcasted_args, downcasted_kwargs)


def upcast(value: Any) -> Any:
    upcast_method = getattr(value, "upcast", None)
    if upcast_method is not None:
        return upcast_method()
    return value


def upcast_args_and_kwargs(
    args: Union[List[Any], Tuple[Any, ...]], kwargs: Dict[Any, Any]
) -> Tuple[List[Any], Dict[Any, Any]]:
    # When we invoke remote methods that use C code and cannot utilise our wrapped
    # types through duck typing, we must convert them to the their original form.
    upcasted_args = []
    upcasted_kwargs = {}
    for arg in args:
        # try to upcast if possible
        upcasted_args.append(upcast(value=arg))

    for k, arg in kwargs.items():
        # try to upcast if possible
        upcasted_kwargs[k] = upcast(value=arg)

    return (upcasted_args, upcasted_kwargs)



# third party
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Float
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String

# relative
from . import Base


class UserApplication(Base):
    __tablename__ = "syft_application"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    email = Column(String(255))
    name = Column(String(255), default="")
    hashed_password = Column(String(512))
    salt = Column(String(255))
    daa_pdf = Column(Integer, ForeignKey("daa_pdf.id"))
    status = Column(String(255), default="pending")
    added_by = Column(String(2048))
    website = Column(String(2048))
    institution = Column(String(2048))
    budget = Column(Float(), default=0.0)

    def __str__(self) -> str:
        return (
            f"<User Application id: {self.id}, email: {self.email}, name: {self.name}"
            f"status: {self.status}>"
        )


class SyftUser(Base):
    __tablename__ = "syft_user"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    email = Column(String(255))
    name = Column(String(255), default="")
    budget = Column(Float(), default=0.0)
    hashed_password = Column(String(512))
    salt = Column(String(255))
    private_key = Column(String(2048))
    verify_key = Column(String(2048))
    role = Column(Integer, ForeignKey("role.id"))
    added_by = Column(String(2048))
    website = Column(String(2048))
    institution = Column(String(2048))
    daa_pdf = Column(Integer, ForeignKey("daa_pdf.id"))
    created_at = Column(DateTime())

    def __str__(self) -> str:
        return (
            f"<User id: {self.id}, email: {self.email}, name: {self.name}"
            f"role: {self.role}>"
        )


def create_user(
    email: str,
    hashed_password: str,
    salt: str,
    private_key: str,
    role: int,
    name: str = "",
    budget: float = 0.0,
) -> SyftUser:
    new_user = SyftUser(
        email=email,
        hashed_password=hashed_password,
        salt=salt,
        private_key=private_key,
        role=role,
        name=name,
        budget=budget,
    )
    return new_user



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/pointer/pointer.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n proto/core/pointer/pointer.proto\x12\x11syft.core.pointer\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\xcd\x02\n\x07Pointer\x12"\n\x1apoints_to_object_with_path\x18\x01 \x01(\t\x12\x14\n\x0cpointer_name\x18\x02 \x01(\t\x12-\n\x0eid_at_location\x18\x03 \x01(\x0b\x32\x15.syft.core.common.UID\x12\'\n\x08location\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0c\n\x04tags\x18\x05 \x03(\t\x12\x13\n\x0b\x64\x65scription\x18\x06 \x01(\t\x12\x13\n\x0bobject_type\x18\x07 \x01(\t\x12\x16\n\x0e\x61ttribute_name\x18\x08 \x01(\t\x12\x19\n\x0cpublic_shape\x18\t \x01(\x0cH\x00\x88\x01\x01\x12\x1e\n\x11obj_public_kwargs\x18\n \x01(\x0cH\x01\x88\x01\x01\x42\x0f\n\r_public_shapeB\x14\n\x12_obj_public_kwargsb\x06proto3'
)


_POINTER = DESCRIPTOR.message_types_by_name["Pointer"]
Pointer = _reflection.GeneratedProtocolMessageType(
    "Pointer",
    (_message.Message,),
    {
        "DESCRIPTOR": _POINTER,
        "__module__": "proto.core.pointer.pointer_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.pointer.Pointer)
    },
)
_sym_db.RegisterMessage(Pointer)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _POINTER._serialized_start = 124
    _POINTER._serialized_end = 457
# @@protoc_insertion_point(module_scope)



# CLEANUP NOTES (for ISHAN):
# - add documentation for each method
# - add comments inline explaining each piece
# - add a unit test for each method (at least)

# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import List as TypeList
from typing import Optional

# third party
from nacl.signing import VerifyKey


# the most generic class
class Scalar:
    """
    A Scalar is the most generic class, which keeps track of the current value, and a data-independent
    min-val and max-val.
    """

    def publish(
        self, acc: Any, user_key: VerifyKey, sigma: float = 1.5
    ) -> TypeList[Any]:
        """Adversarial accountant adds Gaussian noise and publishes the scalar's value"""
        # relative
        from ...publish import publish

        return publish([self], acc=acc, sigma=sigma, user_key=user_key)

    @property
    def max_val(self) -> Optional[float]:
        raise NotImplementedError

    @property
    def min_val(self) -> Optional[float]:
        raise NotImplementedError

    @property
    def value(self) -> Optional[float]:
        raise NotImplementedError

    def __str__(self) -> str:
        return (
            "<"
            + str(type(self).__name__)
            + ": ("
            + str(self.min_val)
            + " < "
            + str(self.value)
            + " < "
            + str(self.max_val)
            + ")>"
        )

    def __repr__(self) -> str:
        return str(self)



# stdlib
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Type
from typing import Union

# third party
from nacl.encoding import HexEncoder
from nacl.signing import SigningKey
from nacl.signing import VerifyKey

# syft absolute
import syft as sy

# relative
from ......grid import GridURL
from ......logger import error
from ......logger import info
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import SignedImmediateSyftMessageWithReply
from ....domain.domain_interface import DomainInterface
from ....domain.enums import AssociationRequestResponses
from ...exceptions import AuthorizationError
from ...exceptions import MissingRequestKeyError
from ...node_service.vpn.vpn_messages import VPNStatusMessageWithReply
from ...node_table.association_request import AssociationRequest
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from ..success_resp_message import SuccessResponseMessage
from .association_request_messages import DeleteAssociationRequestMessage
from .association_request_messages import GetAssociationRequestMessage
from .association_request_messages import GetAssociationRequestResponse
from .association_request_messages import GetAssociationRequestsMessage
from .association_request_messages import GetAssociationRequestsResponse
from .association_request_messages import ReceiveAssociationRequestMessage
from .association_request_messages import RespondAssociationRequestMessage
from .association_request_messages import SendAssociationRequestMessage


def get_vpn_status_metadata(node: DomainInterface) -> Dict[str, Any]:
    vpn_status_msg = (
        VPNStatusMessageWithReply()
        .to(address=node.address, reply_to=node.address)
        .sign(signing_key=node.signing_key)
    )
    vpn_status = node.recv_immediate_msg_with_reply(msg=vpn_status_msg)
    vpn_status_message_contents = vpn_status.message
    status = vpn_status_message_contents.payload.kwargs  # type: ignore
    network_vpn_ip = status["host"]["ip"]
    node_name = status["host"]["hostname"]
    metadata = {
        "host_or_ip": str(network_vpn_ip),
        "node_id": str(node.target_id.id.no_dash),
        "node_name": str(node_name),
        "type": f"{str(type(node).__name__).lower()}",
    }
    return metadata


def check_if_is_vpn(host_or_ip: str) -> bool:
    VPN_IP_SUBNET = "100.64.0."
    return host_or_ip.startswith(VPN_IP_SUBNET)


# domain gets this message from a user and will try to send to the network
def send_association_request_msg(
    msg: SendAssociationRequestMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:
    # Check Key permissions
    info(
        f"Node {node} - send_association_request_msg: got SendAssociationRequestMessage. "
        f"Info: {msg.source} - {msg.target}"
    )
    allowed = node.users.can_manage_infrastructure(verify_key=verify_key)
    if allowed:
        user = node.users.get_user(verify_key=verify_key)
        info(
            f"Node {node} - send_association_request_msg: {node} got user performing the action. User: {user}"
        )

        # Build an association request to send to the target
        user_priv_key = SigningKey(
            node.users.get_user(verify_key).private_key.encode(), encoder=HexEncoder  # type: ignore
        )

        metadata = dict(msg.metadata)
        # get domain metadata to send to the network
        try:
            # TODO: refactor to not stuff our vpn_metadata into the normal metadata
            # because it gets blindly **splatted into the database
            vpn_metadata = get_vpn_status_metadata(node=node)
            metadata.update(vpn_metadata)
        except Exception as e:
            error(f"failed to get vpn status. {e}")

        metadata["node_name"] = (
            node.name if node.name else ""
        )  # tell the network what our name is
        grid_url = GridURL.from_url(msg.target).with_path("/api/v1")
        target_client = sy.connect(url=str(grid_url))

        target_msg: SignedImmediateSyftMessageWithReply = (
            ReceiveAssociationRequestMessage(
                address=target_client.address,
                reply_to=node.address,
                metadata=metadata,
                source=vpn_metadata["host_or_ip"],
                target=msg.target,
            ).sign(signing_key=user_priv_key)
        )

        # Send the message to the target
        info(
            f"Node {node} - send_association_request_msg: sending ReceiveAssociationRequestMessage."
        )
        try:
            # we need target
            target_client.send_immediate_msg_with_reply(msg=target_msg)
        except Exception as e:
            error(f"Failed to send ReceiveAssociationRequestMessage. {e}")

        info(
            f"Node {node} - send_association_request_msg: received the answer from ReceiveAssociationRequestMessage."
        )

        # Create a new association request object
        info(
            f"Node {node} - send_association_request_msg: adding requests to the Database."
        )
        node.association_requests.create_association_request(
            node_name=target_client.name,  # type: ignore
            node_address=target_client.target_id.id.no_dash,  # type: ignore
            status=AssociationRequestResponses.PENDING,
            source=msg.source,
            target=msg.target,
        )
    else:  # If not authorized
        raise AuthorizationError("You're not allowed to create an Association Request!")
    info(f"Node: {node} received the answer from ReceiveAssociationRequestMessage.")
    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="Association request sent!",
    )


# network gets the above message first and then later the domain gets this message as well
def recv_association_request_msg(
    msg: ReceiveAssociationRequestMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:
    _previous_request = node.association_requests.contain(
        source=msg.source, target=msg.target
    )
    info(
        f"Node {node} - recv_association_request_msg: prev request exists {not _previous_request}."
    )

    # Create a new Association Request if the handshake value doesn't exist in the database
    if not _previous_request:
        # this side happens on the network
        info(
            f"Node {node} - recv_association_request_msg: creating a new association request."
        )

        if node.settings.DOMAIN_ASSOCIATION_REQUESTS_AUTOMATICALLY_ACCEPTED:
            status = AssociationRequestResponses.ACCEPT
        else:
            status = AssociationRequestResponses.PENDING

        node.association_requests.create_association_request(
            node_name=msg.metadata["node_name"],
            node_address=msg.reply_to.target_id.id.no_dash,
            status=status,
            source=msg.source,
            target=msg.target,
        )
    else:
        # this side happens on the domain
        info(
            f"Node {node} - recv_association_request_msg: answering an existing association request."
        )
        node.association_requests.set(source=msg.source, target=msg.target, response=msg.response)  # type: ignore

    # get or create a new node and node_route which represents the opposing node which
    # is supplied in the metadata
    try:
        node_id = node.node.create_or_get_node(  # type: ignore
            node_uid=msg.metadata["node_id"], node_name=msg.metadata["node_name"]
        )
        is_vpn = check_if_is_vpn(host_or_ip=msg.metadata["host_or_ip"])
        node.node_route.update_route_for_node(  # type: ignore
            node_id=node_id, host_or_ip=msg.metadata["host_or_ip"], is_vpn=is_vpn
        )
    except Exception as e:
        error(f"Failed to save the node and node_route rows. {e}")

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="Association request received!",
    )


# network owner user approves the request and sends this to the network
def respond_association_request_msg(
    msg: RespondAssociationRequestMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:
    # Check if handshake/address/value fields are empty
    missing_paramaters = not msg.target or not msg.response
    if missing_paramaters:
        raise MissingRequestKeyError(
            message="Invalid request payload, empty fields (target/handshake/value)!"
        )
    # Check Key permissions
    allowed = node.users.can_manage_infrastructure(verify_key=verify_key)

    info(
        f"Node {node} - respond_association_request_msg: user can approve/deny association requests."
    )
    if allowed:
        # Set the status of the Association Request according to the "value" field received

        node.association_requests.set(source=msg.source, target=msg.target, response=msg.response)  # type: ignore
        user_priv_key = SigningKey(
            node.users.get_user(verify_key).private_key.encode(), encoder=HexEncoder  # type: ignore
        )

        metadata = {}
        # get network metadata to send back to domain
        try:
            metadata = get_vpn_status_metadata(node=node)
        except Exception as e:
            error(f"Failed to get vpn status. {e}")

        # create a client to the source
        grid_url = GridURL.from_url(msg.source).with_path("/api/v1")
        source_client = sy.connect(url=str(grid_url))

        try:
            node_msg: SignedImmediateSyftMessageWithReply = (
                ReceiveAssociationRequestMessage(
                    address=source_client.address,
                    response=msg.response,
                    reply_to=node.address,
                    metadata=metadata,
                    source=msg.source,
                    target=msg.target,
                ).sign(signing_key=user_priv_key)
            )

            info(
                f"Node {node} - respond_association_request_msg: sending ReceiveAssociationRequestMessage."
            )

            source_client.send_immediate_msg_with_reply(msg=node_msg)

            info(
                f"Node {node} - respond_association_request_msg: ReceiveAssociationRequestMessage got back."
            )
        except Exception as e:
            error(f"Failed to send ReceiveAssociationRequestMessage to the domain. {e}")

    else:  # If not allowed
        raise AuthorizationError("You're not allowed to create an Association Request!")

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="Association request replied!",
    )


def get_association_request_msg(
    msg: GetAssociationRequestMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> GetAssociationRequestResponse:
    # Check Key Permissions
    allowed = node.users.can_manage_infrastructure(verify_key=verify_key)
    # If allowed
    if allowed:
        association_request: AssociationRequest = node.association_requests.first(
            id=msg.association_id
        )
    else:  # Otherwise
        raise AuthorizationError(
            "You're not allowed to get Association Request information!"
        )

    return GetAssociationRequestResponse(
        address=msg.reply_to,
        content=association_request.get_metadata(),
        source=association_request.source,
        target=association_request.target,
    )


def get_all_association_request_msg(
    msg: GetAssociationRequestsMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> GetAssociationRequestsResponse:
    allowed = node.users.can_manage_infrastructure(verify_key=verify_key)

    # If allowed
    if allowed:
        association_requests = node.association_requests.all()

        association_requests_json = [
            association_request.get_metadata()
            for association_request in association_requests
        ]
    else:  # Otherwise
        raise AuthorizationError(
            "You're not allowed to get Association Request information!"
        )

    return GetAssociationRequestsResponse(
        address=msg.reply_to,
        content=association_requests_json,
    )


def del_association_request_msg(
    msg: DeleteAssociationRequestMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:
    # Check Key permissions
    allowed = node.users.can_manage_infrastructure(
        verify_key=verify_key
    ) and node.association_requests.contain(id=msg.association_id)

    # If allowed
    if allowed:
        node.association_requests.delete(id=msg.association_id)
    else:  # Otherwise
        raise AuthorizationError(
            "You're not allowed to delete this Association Request information!"
        )

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="Association request deleted!",
    )


class AssociationRequestService(ImmediateNodeServiceWithReply):

    msg_handler_map: Dict[type, Callable] = {
        SendAssociationRequestMessage: send_association_request_msg,
        ReceiveAssociationRequestMessage: recv_association_request_msg,
        GetAssociationRequestMessage: get_association_request_msg,
        GetAssociationRequestsMessage: get_all_association_request_msg,
        DeleteAssociationRequestMessage: del_association_request_msg,
        RespondAssociationRequestMessage: respond_association_request_msg,
    }

    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: DomainInterface,
        msg: Union[
            SendAssociationRequestMessage,
            ReceiveAssociationRequestMessage,
            GetAssociationRequestMessage,
            DeleteAssociationRequestMessage,
        ],
        verify_key: VerifyKey,
    ) -> Union[
        SuccessResponseMessage,
        GetAssociationRequestsResponse,
        GetAssociationRequestResponse,
    ]:
        return AssociationRequestService.msg_handler_map[type(msg)](
            msg=msg, node=node, verify_key=verify_key
        )

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithReply]]:
        return [
            SendAssociationRequestMessage,
            ReceiveAssociationRequestMessage,
            GetAssociationRequestMessage,
            GetAssociationRequestsMessage,
            DeleteAssociationRequestMessage,
            RespondAssociationRequestMessage,
        ]



# third party
import pyarrow as pa
import torch as th

# relative
from ...experimental_flags import flags
from ...proto.lib.torch.tensor_pb2 import ProtobufContent
from ...proto.lib.torch.tensor_pb2 import TensorData

# Torch dtypes to string (and back) mappers
TORCH_DTYPE_STR = {
    th.uint8: "uint8",
    th.int8: "int8",
    th.int16: "int16",
    th.int32: "int32",
    th.int64: "int64",
    th.float16: "float16",
    th.float32: "float32",
    th.float64: "float64",
    th.complex32: "complex32",
    th.complex64: "complex64",
    th.complex128: "complex128",
    th.bool: "bool",
    th.qint8: "qint8",
    th.quint8: "quint8",
    th.qint32: "qint32",
    th.bfloat16: "bfloat16",
}
TORCH_STR_DTYPE = {name: cls for cls, name in TORCH_DTYPE_STR.items()}


def protobuf_data_encoding(tensor: th.Tensor) -> bytes:
    protobuf_tensor_data = ProtobufContent()

    if tensor.is_quantized:
        data = th.flatten(tensor).int_repr().tolist()
    else:
        data = th.flatten(tensor).tolist()

    dtype = TORCH_DTYPE_STR[tensor.dtype]
    protobuf_tensor_data.shape.extend(tensor.size())
    getattr(protobuf_tensor_data, "contents_" + dtype).extend(data)
    return protobuf_tensor_data.SerializeToString()


def arrow_data_encoding(tensor: th.Tensor) -> bytes:
    if TORCH_DTYPE_STR[tensor.dtype] == "bfloat16":
        tensor = tensor.type(th.float32)

    if tensor.is_quantized:
        numpy_tensor = tensor.detach().int_repr().numpy()
    else:
        numpy_tensor = tensor.detach().numpy()

    apache_arrow = pa.Tensor.from_numpy(obj=numpy_tensor)
    sink = pa.BufferOutputStream()
    pa.ipc.write_tensor(apache_arrow, sink)
    return sink.getvalue().to_pybytes()


def tensor_serializer(tensor: th.Tensor) -> TensorData:
    """Strategy to serialize a tensor using Protobuf"""

    protobuf_tensor = TensorData()

    if tensor.is_quantized:
        protobuf_tensor.is_quantized = True
        protobuf_tensor.scale = tensor.q_scale()
        protobuf_tensor.zero_point = tensor.q_zero_point()

    if flags.APACHE_ARROW_TENSOR_SERDE:
        protobuf_tensor.arrow_data = arrow_data_encoding(tensor)
    else:
        protobuf_tensor.proto_data = protobuf_data_encoding(tensor)

    protobuf_tensor.dtype = TORCH_DTYPE_STR[tensor.dtype]
    return protobuf_tensor.SerializeToString()


def protobuf_data_decoding(protobuf_tensor: TensorData) -> th.Tensor:
    proto_data = ProtobufContent()
    proto_data.ParseFromString(protobuf_tensor.proto_data)
    size = tuple(proto_data.shape)
    data = getattr(proto_data, "contents_" + protobuf_tensor.dtype)

    if protobuf_tensor.is_quantized:
        # Drop the 'q' from the beginning of the quantized dtype to get the int type
        dtype = TORCH_STR_DTYPE[protobuf_tensor.dtype[1:]]
        int_tensor = th.tensor(data, dtype=dtype).reshape(size)
        # Automatically converts int types to quantized types
        return th._make_per_tensor_quantized_tensor(
            int_tensor, protobuf_tensor.scale, protobuf_tensor.zero_point
        )
    else:
        dtype = TORCH_STR_DTYPE[protobuf_tensor.dtype]
        return th.tensor(data, dtype=dtype).reshape(size)


def arrow_data_decoding(tensor_data: TensorData) -> th.Tensor:
    reader = pa.BufferReader(tensor_data.arrow_data)
    buf = reader.read_buffer()
    result = pa.ipc.read_tensor(buf)
    np_array = result.to_numpy()
    np_array.setflags(write=True)
    data = th.from_numpy(np_array)

    if tensor_data.is_quantized:
        result = th._make_per_tensor_quantized_tensor(
            data, tensor_data.scale, tensor_data.zero_point
        )
    else:
        result = data

    if tensor_data.dtype == "bfloat16":
        result = result.type(th.bfloat16).clone()

    if tensor_data.dtype == "bool":
        result = result.type(th.bool).clone()

    return result


def tensor_deserializer(buf: bytes) -> th.Tensor:
    protobuf_tensor = TensorData()
    protobuf_tensor.ParseFromString(buf)

    if protobuf_tensor.HasField("arrow_data"):
        return arrow_data_decoding(protobuf_tensor)
    elif protobuf_tensor.HasField("proto_data"):
        return protobuf_data_decoding(protobuf_tensor)



# stdlib
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

# third party
from pydantic import BaseSettings

# relative
from ...core.common.serde.deserialize import _deserialize as deserialize
from ...core.common.serde.serializable import serializable
from ...core.common.uid import UID
from ...grid import GridURL


@serializable(recursive_serde=True)
class ProxyDataset:
    __attr_allowlist__ = [
        "node_id",
        "asset_name",
        "dataset_name",
        "shape",
        "dtype",
        "fqn",
        "url",
        "obj_public_kwargs",
    ]

    def __init__(
        self,
        asset_name: str,
        dataset_name: str,
        shape: Tuple[int, ...],
        dtype: str,
        fqn: str,
        node_id: UID,
        obj_public_kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.asset_name = asset_name
        self.dataset_name = dataset_name
        self.shape = shape
        self.dtype = dtype
        self.fqn = fqn
        self.node_id = node_id
        self.url = ""
        self.obj_public_kwargs = (
            obj_public_kwargs if obj_public_kwargs is not None else {}
        )

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__}: {self.dataset_name}[{self.asset_name}] -> "
            + f"shape: {self.shape} dtype: {self.dtype} url: {self.url}>"
        )

    @property
    def name(self) -> str:
        return self.dataset_name + "/" + self.asset_name

    @property
    def data_fully_qualified_name(self) -> str:
        return self.fqn

    def get_s3_data(self, settings: BaseSettings) -> Any:
        try:
            # relative
            from ..node.common.util import get_s3_client

            s3_client = get_s3_client(settings=settings)
            if s3_client is None:
                raise Exception("get_s3_client returned None")
            response = s3_client.get_object(Bucket=self.node_id.no_dash, Key=self.name)
            data = response.get("Body", b"").read()
            return deserialize(data, from_bytes=True)
        except Exception as e:
            print(f"Failed to get data from proxy object {e}.")
            raise e

    def generate_presigned_url(
        self, settings: BaseSettings, public_url: bool = False
    ) -> None:
        # relative
        from ..node.common.util import get_s3_client

        s3_client = get_s3_client(settings=settings)

        download_url = s3_client.generate_presigned_url(
            ClientMethod="get_object",
            Params={"Bucket": self.node_id.no_dash, "Key": self.name},
            ExpiresIn=settings.S3_PRESIGNED_TIMEOUT_SECS,
            HttpMethod="GET",
        )

        if public_url:
            grid_url = GridURL.from_url(url=download_url)
            # add /blob to path
            grid_url.path = f"/blob{grid_url.path}"
            download_url = grid_url.url_path

        self.url = download_url



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/domain/service/request_message.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n4proto/core/node/domain/service/request_message.proto\x12\x1dsyft.core.node.domain.service\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\xdd\x03\n\x0eRequestMessage\x12)\n\nrequest_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12(\n\tobject_id\x18\x02 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bobject_tags\x18\x03 \x03(\t\x12\x1b\n\x13request_description\x18\x04 \x01(\t\x12\x0e\n\x06status\x18\x05 \x01(\t\x12\x14\n\x0crequest_type\x18\x06 \x01(\t\x12\x0c\n\x04\x64\x61te\x18\x07 \x01(\t\x12-\n\x0etarget_address\x18\x08 \x01(\x0b\x32\x15.syft.core.io.Address\x12,\n\rowner_address\x18\t \x01(\x0b\x32\x15.syft.core.io.Address\x12\x1c\n\x14requester_verify_key\x18\n \x01(\x0c\x12\x14\n\x0ctimeout_secs\x18\x0b \x01(\x05\x12\x13\n\x0bobject_type\x18\x0c \x01(\t\x12\x11\n\tuser_name\x18\r \x01(\t\x12\x12\n\nuser_email\x18\x0e \x01(\t\x12\x11\n\tuser_role\x18\x0f \x01(\t\x12\x18\n\x10requested_budget\x18\x10 \x01(\x02\x12\x16\n\x0e\x63urrent_budget\x18\x11 \x01(\x02\x62\x06proto3'
)


_REQUESTMESSAGE = DESCRIPTOR.message_types_by_name["RequestMessage"]
RequestMessage = _reflection.GeneratedProtocolMessageType(
    "RequestMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _REQUESTMESSAGE,
        "__module__": "proto.core.node.domain.service.request_message_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.domain.service.RequestMessage)
    },
)
_sym_db.RegisterMessage(RequestMessage)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _REQUESTMESSAGE._serialized_start = 156
    _REQUESTMESSAGE._serialized_end = 633
# @@protoc_insertion_point(module_scope)



# stdlib
from typing import Optional

# relative
from ...logger import traceback_and_raise
from ..common.message import SignedEventualSyftMessageWithoutReply
from ..common.message import SignedImmediateSyftMessageWithReply
from ..common.message import SignedImmediateSyftMessageWithoutReply


class BidirectionalConnection:
    def recv_immediate_msg_with_reply(
        self, msg: SignedImmediateSyftMessageWithReply
    ) -> SignedImmediateSyftMessageWithoutReply:
        traceback_and_raise(NotImplementedError)

    def recv_immediate_msg_without_reply(
        self, msg: SignedImmediateSyftMessageWithoutReply
    ) -> None:
        traceback_and_raise(NotImplementedError)

    def recv_eventual_msg_without_reply(
        self, msg: SignedEventualSyftMessageWithoutReply
    ) -> None:
        traceback_and_raise(NotImplementedError)

    def send_immediate_msg_with_reply(
        self, msg: SignedImmediateSyftMessageWithReply, timeout: Optional[float] = None
    ) -> SignedImmediateSyftMessageWithoutReply:
        traceback_and_raise(NotImplementedError)

    def send_immediate_msg_without_reply(
        self,
        msg: SignedImmediateSyftMessageWithoutReply,
        timeout: Optional[float] = None,
    ) -> None:
        traceback_and_raise(NotImplementedError)

    def send_eventual_msg_without_reply(
        self,
        msg: SignedEventualSyftMessageWithoutReply,
        timeout: Optional[float] = None,
    ) -> None:
        traceback_and_raise(NotImplementedError)


class ServerConnection:
    def __init__(self) -> None:
        self.opt_bidirectional_conn = BidirectionalConnection()

    def recv_immediate_msg_with_reply(
        self,
        msg: SignedImmediateSyftMessageWithReply,
    ) -> SignedImmediateSyftMessageWithoutReply:
        traceback_and_raise(NotImplementedError)

    def recv_immediate_msg_without_reply(
        self, msg: SignedImmediateSyftMessageWithoutReply
    ) -> None:
        traceback_and_raise(NotImplementedError)

    def recv_eventual_msg_without_reply(
        self, msg: SignedEventualSyftMessageWithoutReply
    ) -> None:
        traceback_and_raise(NotImplementedError)


class ClientConnection:
    def __init__(self) -> None:
        self.opt_bidirectional_conn = BidirectionalConnection()

    def send_immediate_msg_with_reply(
        self, msg: SignedImmediateSyftMessageWithReply, timeout: Optional[float] = None
    ) -> SignedImmediateSyftMessageWithoutReply:
        traceback_and_raise(NotImplementedError)

    def send_immediate_msg_without_reply(
        self,
        msg: SignedImmediateSyftMessageWithoutReply,
        timeout: Optional[float] = None,
    ) -> None:
        traceback_and_raise(NotImplementedError)

    def send_eventual_msg_without_reply(
        self,
        msg: SignedEventualSyftMessageWithoutReply,
        timeout: Optional[float] = None,
    ) -> None:
        traceback_and_raise(NotImplementedError)



# stdlib
from typing import Sequence


def is_broadcastable(shape1: Sequence, shape2: Sequence) -> bool:
    """Helper function to determine if Tensor Operations can be broadcast
    inputs:
    shape1, shape 2: shapes of numpy arrays/syft tensors

    outputs:
    True or False
    """
    for a, b in zip(shape1[::-1], shape2[::-1]):
        if a == 1 or b == 1 or a == b:
            pass
        else:
            return False
    return True


# def is_mat_multiplicable(shape1: Sequence, shape2: Sequence) -> bool:
#     if len(shape1) >= 2:
#         if shape1[-1] == shape2[-2]:
#             return True
#         return False
#     else:
#         if shape1[0] == shape2[-2]:
#             return True
#         return False



# stdlib
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Tuple as TypeTuple

HANDLED_FUNCTIONS: Dict[Any, Any] = {}


def inputs2child(*args: Tuple[Any, ...], **kwargs: Any) -> TypeTuple[Tuple[Any], Any]:

    # relative
    from .passthrough import PassthroughTensor  # type: ignore

    out_args_list = list()

    for out in tuple(
        [x.child if isinstance(x, PassthroughTensor) else x for x in args]
    ):
        out_args_list.append(out)

    out_kwargs = {}

    for x in kwargs.items():
        if isinstance(x[1], PassthroughTensor):
            out_kwargs[x[0]] = x[1].child
        else:
            out_kwargs[x[0]] = x[1]

    out_args = tuple(out_args_list)

    return out_args, out_kwargs  # type: ignore


def query_implementation(tensor_type: Any, func: Any) -> Any:
    name = func.__name__
    cache = HANDLED_FUNCTIONS.get(tensor_type, None)
    if cache and name in cache:
        return HANDLED_FUNCTIONS[tensor_type][func.__name__]
    return None


def implements(tensor_type: Any, np_function: Any) -> Any:
    def decorator(func: Any) -> Any:
        if tensor_type not in HANDLED_FUNCTIONS:
            HANDLED_FUNCTIONS[tensor_type] = {}

        HANDLED_FUNCTIONS[tensor_type][np_function.__name__] = func
        return func

    return decorator



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/messages/setup_messages.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n(proto/grid/messages/setup_messages.proto\x12\x12syft.grid.messages\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\xe7\x01\n\x19\x43reateInitialSetUpMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\r\n\x05\x65mail\x18\x04 \x01(\t\x12\x10\n\x08password\x18\x05 \x01(\t\x12\x13\n\x0b\x64omain_name\x18\x06 \x01(\t\x12\x0e\n\x06\x62udget\x18\x07 \x01(\x01\x12\'\n\x08reply_to\x18\x08 \x01(\x0b\x32\x15.syft.core.io.Address"\x89\x01\n\x0fGetSetUpMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"\x87\x01\n\x10GetSetUpResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\xf8\x01\n\x12UpdateSetupMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x13\n\x0b\x64omain_name\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x0b\n\x03\x64\x61\x61\x18\x05 \x01(\x08\x12\x0f\n\x07\x63ontact\x18\x06 \x01(\t\x12\x14\n\x0c\x64\x61\x61_document\x18\x08 \x01(\x0c\x12\x0c\n\x04tags\x18\t \x03(\t\x12\'\n\x08reply_to\x18\n \x01(\x0b\x32\x15.syft.core.io.Address"u\n\x13UpdateSetupResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)


_CREATEINITIALSETUPMESSAGE = DESCRIPTOR.message_types_by_name[
    "CreateInitialSetUpMessage"
]
_GETSETUPMESSAGE = DESCRIPTOR.message_types_by_name["GetSetUpMessage"]
_GETSETUPRESPONSE = DESCRIPTOR.message_types_by_name["GetSetUpResponse"]
_UPDATESETUPMESSAGE = DESCRIPTOR.message_types_by_name["UpdateSetupMessage"]
_UPDATESETUPRESPONSE = DESCRIPTOR.message_types_by_name["UpdateSetupResponse"]
CreateInitialSetUpMessage = _reflection.GeneratedProtocolMessageType(
    "CreateInitialSetUpMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATEINITIALSETUPMESSAGE,
        "__module__": "proto.grid.messages.setup_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.CreateInitialSetUpMessage)
    },
)
_sym_db.RegisterMessage(CreateInitialSetUpMessage)

GetSetUpMessage = _reflection.GeneratedProtocolMessageType(
    "GetSetUpMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETSETUPMESSAGE,
        "__module__": "proto.grid.messages.setup_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetSetUpMessage)
    },
)
_sym_db.RegisterMessage(GetSetUpMessage)

GetSetUpResponse = _reflection.GeneratedProtocolMessageType(
    "GetSetUpResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETSETUPRESPONSE,
        "__module__": "proto.grid.messages.setup_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetSetUpResponse)
    },
)
_sym_db.RegisterMessage(GetSetUpResponse)

UpdateSetupMessage = _reflection.GeneratedProtocolMessageType(
    "UpdateSetupMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _UPDATESETUPMESSAGE,
        "__module__": "proto.grid.messages.setup_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.UpdateSetupMessage)
    },
)
_sym_db.RegisterMessage(UpdateSetupMessage)

UpdateSetupResponse = _reflection.GeneratedProtocolMessageType(
    "UpdateSetupResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _UPDATESETUPRESPONSE,
        "__module__": "proto.grid.messages.setup_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.UpdateSetupResponse)
    },
)
_sym_db.RegisterMessage(UpdateSetupResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _CREATEINITIALSETUPMESSAGE._serialized_start = 133
    _CREATEINITIALSETUPMESSAGE._serialized_end = 364
    _GETSETUPMESSAGE._serialized_start = 367
    _GETSETUPMESSAGE._serialized_end = 504
    _GETSETUPRESPONSE._serialized_start = 507
    _GETSETUPRESPONSE._serialized_end = 642
    _UPDATESETUPMESSAGE._serialized_start = 645
    _UPDATESETUPMESSAGE._serialized_end = 893
    _UPDATESETUPRESPONSE._serialized_start = 895
    _UPDATESETUPRESPONSE._serialized_end = 1012
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/numpy/array.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1bproto/lib/numpy/array.proto\x12\x0esyft.lib.numpy"j\n\nNumpyProto\x12\x14\n\narrow_data\x18\x01 \x01(\x0cH\x00\x12\x14\n\nproto_data\x18\x02 \x01(\x0cH\x00\x12\r\n\x05\x64type\x18\x03 \x01(\t\x12\x19\n\x11\x64\x65\x63ompressed_size\x18\x04 \x01(\x03\x42\x06\n\x04\x64\x61tab\x06proto3'
)


_NUMPYPROTO = DESCRIPTOR.message_types_by_name["NumpyProto"]
NumpyProto = _reflection.GeneratedProtocolMessageType(
    "NumpyProto",
    (_message.Message,),
    {
        "DESCRIPTOR": _NUMPYPROTO,
        "__module__": "proto.lib.numpy.array_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.numpy.NumpyProto)
    },
)
_sym_db.RegisterMessage(NumpyProto)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _NUMPYPROTO._serialized_start = 47
    _NUMPYPROTO._serialized_end = 153
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/client.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import location_pb2 as proto_dot_core_dot_io_dot_location__pb2
from syft.proto.core.io import route_pb2 as proto_dot_core_dot_io_dot_route__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b"\n#proto/core/node/common/client.proto\x12\x15syft.core.node.common\x1a%proto/core/common/common_object.proto\x1a\x1cproto/core/io/location.proto\x1a\x19proto/core/io/route.proto\"\xee\x02\n\x06\x43lient\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x0c\n\x04name\x18\x02 \x01(\t\x12'\n\x06routes\x18\x03 \x03(\x0b\x32\x17.syft.core.io.SoloRoute\x12\x34\n\x07network\x18\x04 \x01(\x0b\x32\x1e.syft.core.io.SpecificLocationH\x00\x88\x01\x01\x12\x33\n\x06\x64omain\x18\x05 \x01(\x0b\x32\x1e.syft.core.io.SpecificLocationH\x01\x88\x01\x01\x12\x33\n\x06\x64\x65vice\x18\x06 \x01(\x0b\x32\x1e.syft.core.io.SpecificLocationH\x02\x88\x01\x01\x12/\n\x02vm\x18\x07 \x01(\x0b\x32\x1e.syft.core.io.SpecificLocationH\x03\x88\x01\x01\x12\x10\n\x08obj_type\x18\x08 \x01(\tB\n\n\x08_networkB\t\n\x07_domainB\t\n\x07_deviceB\x05\n\x03_vmb\x06proto3"
)


_CLIENT = DESCRIPTOR.message_types_by_name["Client"]
Client = _reflection.GeneratedProtocolMessageType(
    "Client",
    (_message.Message,),
    {
        "DESCRIPTOR": _CLIENT,
        "__module__": "proto.core.node.common.client_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.common.Client)
    },
)
_sym_db.RegisterMessage(Client)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _CLIENT._serialized_start = 159
    _CLIENT._serialized_end = 525
# @@protoc_insertion_point(module_scope)



# third party
from sqlalchemy.engine import Engine

# relative
from ..node_table.node import Node
from .database_manager import DatabaseManager


class NodeManager(DatabaseManager):
    schema = Node

    def __init__(self, database: Engine) -> None:
        super().__init__(schema=NodeManager.schema, db=database)

    def create_or_get_node(self, node_uid: str, node_name: str) -> int:
        # node_uid is a UID as a string with no_dash
        node = self.first(node_uid=node_uid)
        if node:
            self.modify(query={"node_uid": node_uid}, values={"node_name": node_name})
        else:
            self.register(**{"node_uid": node_uid, "node_name": node_name})
            node = self.first(node_uid=node_uid)
        return node.id



# stdlib
from typing import List
from typing import Optional
from typing import Type

# third party
from nacl.signing import VerifyKey

# relative
from ....abstract.node_service_interface import NodeServiceInterface
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from .user_auth_messages import UserLoginMessage
from .user_auth_messages import UserLoginMessageWithReply
from .user_auth_messages import UserLoginReplyMessage


class UserLoginService(ImmediateNodeServiceWithReply):
    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: NodeServiceInterface,
        msg: UserLoginMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> UserLoginReplyMessage:
        # this service requires no verify_key because its currently public
        result = msg.payload.run(node=node, verify_key=verify_key)
        return UserLoginMessageWithReply(kwargs=result).back_to(address=msg.reply_to)

    @staticmethod
    def message_handler_types() -> List[Type[UserLoginMessage]]:
        return [UserLoginMessage]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/io/address.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.io import location_pb2 as proto_dot_core_dot_io_dot_location__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1bproto/core/io/address.proto\x12\x0csyft.core.io\x1a\x1cproto/core/io/location.proto"\xb3\x02\n\x07\x41\x64\x64ress\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08obj_type\x18\x02 \x01(\t\x12\x13\n\x0bhas_network\x18\x03 \x01(\x08\x12/\n\x07network\x18\x04 \x01(\x0b\x32\x1e.syft.core.io.SpecificLocation\x12\x12\n\nhas_domain\x18\x05 \x01(\x08\x12.\n\x06\x64omain\x18\x06 \x01(\x0b\x32\x1e.syft.core.io.SpecificLocation\x12\x12\n\nhas_device\x18\x07 \x01(\x08\x12.\n\x06\x64\x65vice\x18\x08 \x01(\x0b\x32\x1e.syft.core.io.SpecificLocation\x12\x0e\n\x06has_vm\x18\t \x01(\x08\x12*\n\x02vm\x18\n \x01(\x0b\x32\x1e.syft.core.io.SpecificLocationb\x06proto3'
)


_ADDRESS = DESCRIPTOR.message_types_by_name["Address"]
Address = _reflection.GeneratedProtocolMessageType(
    "Address",
    (_message.Message,),
    {
        "DESCRIPTOR": _ADDRESS,
        "__module__": "proto.core.io.address_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.io.Address)
    },
)
_sym_db.RegisterMessage(Address)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _ADDRESS._serialized_start = 76
    _ADDRESS._serialized_end = 383
# @@protoc_insertion_point(module_scope)



# third party
from sqlalchemy.engine import Engine

# relative
from ..node_table.node_route import NodeRoute
from .database_manager import DatabaseManager


class NodeRouteManager(DatabaseManager):
    schema = NodeRoute

    def __init__(self, database: Engine) -> None:
        super().__init__(schema=NodeRouteManager.schema, db=database)

    def update_route_for_node(
        self, node_id: int, host_or_ip: str, is_vpn: bool = False
    ) -> NodeRoute:
        # node_id is a database int id
        # host_or_ip can have a port as well

        node_route = self.first(host_or_ip=host_or_ip)
        if node_route:
            self.modify(
                query={"host_or_ip": host_or_ip},
                values={"is_vpn": is_vpn, "node_id": node_id},
            )
        else:
            self.register(
                **{"node_id": node_id, "host_or_ip": host_or_ip, "is_vpn": is_vpn}
            )
            node_route = self.first(host_or_ip=host_or_ip)
        return node_route



# stdlib
from typing import Dict
from typing import List
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from typing_extensions import final

# relative
from ...... import serialize
from ......proto.grid.messages.dataset_messages_pb2 import (
    CreateDatasetMessage as CreateDatasetMessage_PB,
)
from ......proto.grid.messages.dataset_messages_pb2 import (
    DeleteDatasetMessage as DeleteDatasetMessage_PB,
)
from ......proto.grid.messages.dataset_messages_pb2 import (
    GetDatasetMessage as GetDatasetMessage_PB,
)
from ......proto.grid.messages.dataset_messages_pb2 import (
    GetDatasetResponse as GetDatasetResponse_PB,
)
from ......proto.grid.messages.dataset_messages_pb2 import (
    GetDatasetsMessage as GetDatasetsMessage_PB,
)
from ......proto.grid.messages.dataset_messages_pb2 import (
    GetDatasetsResponse as GetDatasetsResponse_PB,
)
from ......proto.grid.messages.dataset_messages_pb2 import (
    UpdateDatasetMessage as UpdateDatasetMessage_PB,
)
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.deserialize import _deserialize
from .....common.serde.serializable import serializable
from .....common.uid import UID
from .....io.address import Address


@final
@serializable()
class CreateDatasetMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        dataset: bytes,
        metadata: Dict[str, str],
        reply_to: Address,
        platform: str,
        msg_id: Optional[UID] = None,
    ) -> None:
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.dataset = dataset
        self.metadata = metadata
        self.platform = platform

    def _object2proto(self) -> CreateDatasetMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: CreateDatasetMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return CreateDatasetMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            dataset=self.dataset,
            metadata=self.metadata,
            reply_to=serialize(self.reply_to),
            platform=self.platform,
        )

    @staticmethod
    def _proto2object(
        proto: CreateDatasetMessage_PB,
    ) -> "CreateDatasetMessage":
        """Creates a CreateDatasetMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: CreateDatasetMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return CreateDatasetMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            dataset=proto.dataset,
            metadata=dict(proto.metadata),
            reply_to=_deserialize(blob=proto.reply_to),
            platform=proto.platform,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return CreateDatasetMessage_PB


@final
@serializable()
class GetDatasetMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        dataset_id: str,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.dataset_id = dataset_id

    def _object2proto(self) -> GetDatasetMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetDatasetMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetDatasetMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            dataset_id=self.dataset_id,
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetDatasetMessage_PB,
    ) -> "GetDatasetMessage":
        """Creates a GetDatasetMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetDatasetMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetDatasetMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            dataset_id=proto.dataset_id,
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetDatasetMessage_PB


@final
@serializable()
class GetDatasetResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        metadata: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.metadata = metadata

    def _object2proto(self) -> GetDatasetResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetDatasetResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            metadata=self.metadata,
        )

    @staticmethod
    def _proto2object(
        proto: GetDatasetResponse_PB,
    ) -> "GetDatasetResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetDatasetResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            metadata=dict(proto.metadata),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetDatasetResponse_PB


@final
@serializable()
class GetDatasetsMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)

    def _object2proto(self) -> GetDatasetsMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetDatasetsMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetDatasetsMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetDatasetsMessage_PB,
    ) -> "GetDatasetsMessage":
        """Creates a GetDatasetsMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetDatasetsMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetDatasetsMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetDatasetsMessage_PB


@final
@serializable()
class GetDatasetsResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        metadatas: List[Dict],
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.metadatas = metadatas

    def _object2proto(self) -> GetDatasetsResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """

        msg = GetDatasetsResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
        )

        metadata_container = GetDatasetsResponse_PB.metadata_container

        for _metadata in self.metadatas:
            metadata = {}
            for k, v in _metadata.items():
                if not isinstance(v, bytes):
                    metadata[k] = serialize(v, to_bytes=True)
                else:
                    metadata[k] = v

            cm = metadata_container(metadata=metadata)
            msg.metadatas.append(cm)

        return msg

    @staticmethod
    def _proto2object(
        proto: GetDatasetsResponse_PB,
    ) -> "GetDatasetsResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetDatasetsResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            metadatas=[
                dict(metadata_container.metadata)
                for metadata_container in proto.metadatas
            ],
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetDatasetsResponse_PB


@final
@serializable()
class UpdateDatasetMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        dataset_id: str,
        metadata: Dict,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.metadata = metadata
        self.dataset_id = dataset_id

    def _object2proto(self) -> UpdateDatasetMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: UpdateDatasetMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return UpdateDatasetMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            dataset_id=self.dataset_id,
            metadata=self.metadata,
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: UpdateDatasetMessage_PB,
    ) -> "UpdateDatasetMessage":
        """Creates a UpdateDatasetMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: UpdateDatasetMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return UpdateDatasetMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            dataset_id=proto.dataset_id,
            metadata=dict(proto.metadata),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return UpdateDatasetMessage_PB


@final
@serializable()
class DeleteDatasetMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        dataset_id: str,
        reply_to: Address,
        bin_object_id: Optional[str] = None,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.dataset_id = dataset_id
        self.bin_object_id = bin_object_id

    def _object2proto(self) -> DeleteDatasetMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: DeleteDatasetMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return DeleteDatasetMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            dataset_id=self.dataset_id,
            bin_object_id=self.bin_object_id,
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: DeleteDatasetMessage_PB,
    ) -> "DeleteDatasetMessage":
        """Creates a DeleteDatasetMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: DeleteDatasetMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return DeleteDatasetMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            dataset_id=proto.dataset_id,
            bin_object_id=proto.bin_object_id,
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return DeleteDatasetMessage_PB



# type: ignore
# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple as TypeTuple
from typing import Type
from typing import Union

# third party
import numpy as np
import torch

# relative
from .util import implements
from .util import query_implementation

AcceptableSimpleType = Union[int, bool, float, np.ndarray]
SupportedChainType = Union["PassthroughTensor", AcceptableSimpleType]


def is_acceptable_simple_type(obj):
    return isinstance(obj, (int, bool, float, np.ndarray, torch.Tensor))


class PassthroughTensor(np.lib.mixins.NDArrayOperatorsMixin):
    """A simple tensor class which passes method/function calls to self.child"""

    def __init__(self, child: Any) -> None:
        self.child = child

    # TODO: Remove
    @property
    def _data_child(self) -> Any:
        data = self
        while hasattr(data, "child"):
            data = data.child
        return data

    def __len__(self) -> int:
        if (
            isinstance(self.child, float)
            or isinstance(self.child, int)
            or isinstance(self.child, bool)
        ):
            return 1
        return len(self.child)

    @property
    def shape(self) -> TypeTuple[Any, ...]:
        if (
            isinstance(self.child, float)
            or isinstance(self.child, int)
            or isinstance(self.child, bool)
        ):
            return (1,)
        return tuple(self.child.shape)

    @property
    def dtype(self) -> np.dtype:
        return self.child.dtype

    # @property
    # def shape(self) -> Union[TypeTuple[Any, ...], List[Any]]:
    #     """There are 3 options for where shape information can be sourced from:

    #     - self.client_shape: which is a logical attempt on the client side to infer
    #       shape information
    #     - self.child.shape: which just says that this layer of logical abstraction isn't
    #       responsible for tracking shape and inhereits it from the child
    #     - Pointer.shape: a special case of the other two wherein shape is called on a
    #       Pointer and it fetches that information from a remote object if
    #       Pointer.client_shape is not available (is None).

    #     By default the priority for where shape information should come from is:
    #     - self.client_shape is first because this doesn't leverage private information.
    #     - self.child.shape is second as a delegation of the first
    #     - Pointer.client_shape as a special case of the previous
    #     - Pointer.remote_shape is a last resort because it requires calling .request()
    #       and .get()

    #     """
    #     if self.client_shape is not None:
    #         return self.client_shape

    #     elif self.child is not None:

    #         # note that this will attempt client_shape on the child before
    #         # anything else because it too will call this method
    #         shape = self.child.shape

    #         if isinstance(shape, Pointer):
    #             return shape.get(request_block=True)
    #         else:
    #             return shape

    #     else:
    #         msg = (
    #             "Not sure how to find shape because self.client_shape and self.child"
    #             + "are both none"
    #         )
    #         raise Exception(msg)

    #     return tuple(self.child.shape)

    def __and__(self, other):
        if is_acceptable_simple_type(other):
            return self.__class__(self.child & other)
        return self.__class__(self.child & other.child)

    def __rand__(self, other):
        if is_acceptable_simple_type(other):
            return self.__class__(other & self.child)

        return other.__class__(other.child & self.child)

    def __abs__(self) -> Union[Type[PassthroughTensor], AcceptableSimpleType]:
        return self.__class__(self.child.__abs__())

    def __add__(
        self, other: Union[Type[PassthroughTensor], AcceptableSimpleType]
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child + other)
        return self.__class__(self.child + other.child)

    def __radd__(self, other) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(other + self.child)

        return other.__class__(other.child + self.child)

    def __sub__(
        self, other: Union[Type[PassthroughTensor], AcceptableSimpleType]
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child - other)
        return self.__class__(self.child - other.child)

    def __rsub__(self, other) -> PassthroughTensor:
        return self.__class__(-(self - other).child)

    def __gt__(
        self,
        other: Union[Type[PassthroughTensor], AcceptableSimpleType],
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child > other)

        return self.__class__(self.child > other.child)

    def __ge__(
        self,
        other: Union[Type[PassthroughTensor], AcceptableSimpleType],
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child >= other)

        return self.__class__(self.child >= other.child)

    def __lt__(
        self,
        other: Union[Type[PassthroughTensor], AcceptableSimpleType],
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child < other)

        return self.__class__(self.child < other.child)

    def __le__(
        self,
        other: Union[Type[PassthroughTensor], AcceptableSimpleType],
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child <= other)

        return self.__class__(self.child <= other.child)

    def __ne__(
        self,
        other: Union[Type[PassthroughTensor], AcceptableSimpleType],
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child != other)

        return self.__class__(self.child != other.child)

    def __eq__(
        self,
        other: Union[Type[PassthroughTensor], AcceptableSimpleType],
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child == other)

        return self.__class__(self.child == other.child)

    def __floordiv__(
        self,
        other: Union[Type[PassthroughTensor], AcceptableSimpleType],
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child.__floordiv__(other))

        return self.__class__(self.child.__floordiv__(other.child))

    def __rfloordiv__(
        self,
        other: Union[Type[PassthroughTensor], AcceptableSimpleType],
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(other.__floordiv__(self.child))

        return self.__class__(other.child.__floordiv__(self.child))

    def __lshift__(
        self,
        other: Union[Type[PassthroughTensor], AcceptableSimpleType],
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child.__lshift__(other))

        return self.__class__(self.child.__lshift__(other.child))

    def __rlshift__(
        self,
        other: Union[Type[PassthroughTensor], AcceptableSimpleType],
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(other.__lshift__(self.child))

        return self.__class__(other.child.__lshift__(self.child))

    def __rshift__(
        self,
        other: Union[Type[PassthroughTensor], AcceptableSimpleType],
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child.__rshift__(other))

        return self.__class__(self.child.__rshift__(other.child))

    def __rrshift__(
        self,
        other: Union[Type[PassthroughTensor], AcceptableSimpleType],
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(other.__rshift__(self.child))

        return self.__class__(other.child.__rshift__(self.child))

    def __pow__(
        self,
        other: Union[Type[PassthroughTensor], AcceptableSimpleType],
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child.__pow__(other))

        return self.__class__(self.child.__pow__(other.child))

    def __rpow__(
        self, other: Union[Type[PassthroughTensor], AcceptableSimpleType]
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child.__rpow__(other))
        return self.__class__(self.child.__rpow__(other.child))

    def __divmod__(
        self,
        other: Union[Type[PassthroughTensor], AcceptableSimpleType],
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child.__divmod__(other))

        return self.__class__(self.child.__divmod__(other.child))

    def __neg__(self) -> PassthroughTensor:
        return self * -1

    def __index__(self) -> int:
        return self.child.__index__()

    def __invert__(self) -> PassthroughTensor:
        return self.child.__invert__()

    def copy(self, order: Optional[str] = "K") -> PassthroughTensor:
        return self.__class__(self.child.copy(order=order))

    def __mul__(
        self, other: Union[Type[PassthroughTensor], AcceptableSimpleType]
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child * other)

        return self.__class__(self.child * other.child)

    def __rmul__(
        self, other: Union[Type[PassthroughTensor], AcceptableSimpleType]
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child.__rmul__(other))
        return self.__class__(self.child.__rmul__(other.child))

    def __matmul__(
        self, other: Union[Type[PassthroughTensor], np.ndarray]
    ) -> PassthroughTensor:
        return self.manual_dot(other)

    def __rmatmul__(
        self, other: Union[Type[PassthroughTensor], np.ndarray]
    ) -> PassthroughTensor:
        return other.manual_dot(self)

    def __truediv__(
        self, other: Union[Type[PassthroughTensor], AcceptableSimpleType]
    ) -> PassthroughTensor:
        if is_acceptable_simple_type(other):
            return self.__class__(self.child * (1 / other))  # type: ignore

        return self.__class__(self.child / other.child)

    def __rtruediv__(self, other: Type[PassthroughTensor]) -> PassthroughTensor:
        return other.__truediv__(self)

    def manual_dot(
        self, other: Union[Type[PassthroughTensor], np.ndarray]
    ) -> PassthroughTensor:

        expanded_self = self.repeat(other.shape[1], axis=1)
        expanded_self = expanded_self.reshape(
            self.shape[0], self.shape[1], other.shape[1]
        )
        expanded_other = other.reshape([1] + list(other.shape)).repeat(
            self.shape[0], axis=0
        )

        prod = expanded_self * expanded_other
        result = prod.sum(axis=1)
        return result

    def dot(
        self, other: Union[Type[PassthroughTensor], np.ndarray]
    ) -> PassthroughTensor:
        return self.manual_dot(other)

    def reshape(self, *dims) -> PassthroughTensor:
        return self.__class__(self.child.reshape(*dims))

    def repeat(
        self, repeats: Union[int, TypeTuple[int, ...]], axis: Optional[int] = None
    ) -> PassthroughTensor:
        return self.__class__(self.child.repeat(repeats, axis))

    def resize(
        self,
        new_shape: Union[int, TypeTuple[int, ...]],
        refcheck: Optional[bool] = True,
    ) -> PassthroughTensor:
        # Should be modified to  remove copy
        # https://stackoverflow.com/questions/23253144/numpy-the-array-doesnt-have-its-own-data
        res = self.child.copy()
        res.resize(new_shape, refcheck=refcheck)
        return self.__class__(res)

    @property
    def T(self) -> PassthroughTensor:
        return self.transpose()

    def transpose(self, *args, **kwargs):
        return self.__class__(self.child.transpose(*args, **kwargs))

    def __getitem__(
        self, key: Union[int, bool, np.array, PassthroughTensor, slice, Ellipsis]
    ) -> Union[PassthroughTensor, AcceptableSimpleType]:
        if isinstance(key, PassthroughTensor):
            return self.__class__(self.child.__getitem__(key.child))

        return self.__class__(self.child.__getitem__(key))

    # numpy.argmax(a, axis=None, out=None)
    def argmax(self, axis: Optional[int]) -> PassthroughTensor:
        return self.__class__(self.child.argmax(axis))

    def argmin(self, axis: Optional[int]) -> PassthroughTensor:
        return self.__class__(self.child.argmin(axis))

    # numpy.argsort(a, axis=-1, kind=None, order=None)
    def argsort(self, axis: Optional[int] = -1) -> PassthroughTensor:
        return self.__class__(self.child.argsort(axis))

    # ndarray.sort(axis=-1, kind=None, order=None)
    def sort(self, axis: int = -1, kind: Optional[str] = None) -> PassthroughTensor:
        return self.__class__(self.child.sort(axis=axis, kind=kind))

    # numpy.clip(a, a_min, a_max, out=None, **kwargs)
    def clip(
        self,
        min: Optional[AcceptableSimpleType] = None,
        max: Optional[AcceptableSimpleType] = None,
    ) -> PassthroughTensor:
        return self.__class__(self.child.clip(min, max))

    # numpy.cumprod(a, axis=None, dtype=None, out=None)
    def cumprod(self, axis: Optional[int] = None) -> PassthroughTensor:
        return self.__class__(self.child.cumprod(axis=axis))

    # TODO : override default datatype as np.int32 until we have support np.int64
    # numpy.cumsum(a, axis=None, dtype=None, out=None)
    def cumsum(
        self,
        axis: Optional[int] = None,
        dtype: Optional[np.dtype] = np.int32,
        out: Optional[np.ndarray] = None,
    ) -> PassthroughTensor:
        return self.__class__(self.child.cumsum(axis=axis, dtype=dtype, out=out))

    # numpy.trace(a, offset=0, axis1=0, axis2=1, dtype=None, out=None)
    def trace(
        self,
        offset: Optional[int] = 0,
        axis1: Optional[int] = 0,
        axis2: Optional[int] = 1,
        dtype: Optional[np.dtype] = np.int32,
        out: Optional[np.ndarray] = None,
    ) -> PassthroughTensor:
        return self.__class__(
            self.child.trace(
                offset=offset, axis1=axis1, axis2=axis2, dtype=dtype, out=out
            )
        )

    # numpy.diagonal(a, offset=0, axis1=0, axis2=1)
    def diagonal(
        self, offset: int = 0, axis1: int = 0, axis2: int = 1
    ) -> PassthroughTensor:
        return self.__class__(self.child.diagonal(offset, axis1, axis2))

    def tolist(self) -> PassthroughTensor:
        return self.__class__(self.child.tolist())

    # ndarray.flatten(order='C')
    def flatten(self, order: Optional[str] = "C") -> PassthroughTensor:
        return self.__class__(self.child.flatten(order))

    # ndarray.partition(kth, axis=- 1, kind='introselect', order=None)
    def partition(
        self,
        kth: Union[int, TypeTuple[int, ...]],
        axis: Optional[int] = -1,
        kind: Optional[str] = "introselect",
        order: Optional[Union[int, TypeTuple[int, ...]]] = None,
    ) -> PassthroughTensor:
        self.child.partition(kth=kth, axis=axis, kind=kind, order=order)  # inplace op
        return self.__class__(self.child)

    # ndarray.ravel([order])
    def ravel(self, order: Optional[str] = "C") -> PassthroughTensor:
        return self.__class__(self.child.ravel(order=order))

    # ndarray.compress(condition, axis=None, out=None)
    def compress(
        self, condition: List[bool], axis: int = None, out: Optional[np.ndarray] = None
    ) -> PassthroughTensor:
        return self.__class__(
            self.child.compress(condition=condition, axis=axis, out=out)
        )

    # ndarray.swapaxes(axis1, axis2)
    def swapaxes(self, axis1: int, axis2: int) -> PassthroughTensor:
        return self.__class__(self.child.swapaxes(axis1, axis2))

    # ndarray.put(indices, values, mode='raise')
    def put(
        self,
        indices: Union[int, TypeTuple[int, ...], np.ndarray],
        values: Union[int, TypeTuple[int, ...], np.ndarray],
        mode: Optional[str] = "raise",
    ) -> None:
        self.child.put(indices=indices, values=values, mode=mode)  # inplace op
        print(self.__class__)
        print(self.child)
        return self.__class__(self.child)

    # ndarray.__pos__(/)
    def __pos__(self) -> PassthroughTensor:
        return self.__class__(self.child.__pos__())

    # numpy.mean(a, axis=None, dtype=None, out=None, keepdims=<no value>, *, where=<no value>)
    def mean(
        self, axis: Optional[Union[int, TypeTuple[int, ...]]] = None, **kwargs
    ) -> Union[PassthroughTensor, np.floating]:
        return self.__class__(self.child.mean(axis=axis, **kwargs))

    # ndarray.max(axis=None, out=None, keepdims=False, initial=<no value>, where=True)
    def max(
        self, axis: Optional[Union[int, TypeTuple[int, ...]]] = None
    ) -> PassthroughTensor:
        return self.__class__(self.child.max(axis=axis))

    # ndarray.min(axis=None, out=None, keepdims=False, initial=<no value>, where=True)
    def min(
        self, axis: Optional[Union[int, TypeTuple[int, ...]]] = None
    ) -> PassthroughTensor:
        return self.__class__(self.child.min(axis=axis))

    @property
    def ndim(self) -> np.unsignedinteger:
        return self.child.ndim

    # ndarray.prod(axis=None, dtype=None, out=None, keepdims=False, initial=1, where=True)
    def prod(
        self, axis: Optional[Union[int, TypeTuple[int, ...]]] = None
    ) -> PassthroughTensor:
        return self.__class__(self.child.prod(axis=axis))

    # numpy.squeeze(a, axis=None)
    def squeeze(
        self, axis: Optional[Union[int, TypeTuple[int, ...]]] = None
    ) -> PassthroughTensor:
        return self.__class__(self.child.squeeze(axis=axis))

    # numpy.std(a, axis=None, dtype=None, out=None, ddof=0, keepdims=<no value>, *, where=<no value>)
    def std(
        self, axis: Optional[Union[int, TypeTuple[int, ...]]] = None
    ) -> PassthroughTensor:
        return self.__class__(self.child.std(axis=axis))

    # numpy.sum(a, axis=None, dtype=None, out=None, keepdims=<no value>, initial=<no value>, where=<no value>)
    def sum(
        self, axis: Optional[Union[int, TypeTuple[int, ...]]] = None
    ) -> PassthroughTensor:
        result = self.child.sum(axis=axis)
        if hasattr(self, "copy_tensor"):
            tensor = self.copy_tensor()
            tensor.child = result
            return tensor
        return self.__class__(result)

    # numpy.take(a, indices, axis=None, out=None, mode='raise')
    def take(
        self,
        indices: Union[int, TypeTuple[int, ...], np.ndarray],
        axis: Optional[int] = None,
        out: Optional[np.ndarray] = None,
        mode: Optional[str] = "raise",
    ) -> PassthroughTensor:
        return self.__class__(
            self.child.take(
                indices,
                axis=axis,
                out=out,
                mode=mode,
            )
        )

    # numpy.choose(a, choices, out=None, mode='raise')
    def choose(
        self,
        choices: Sequence[Union[PassthroughTensor, np.ndarray]],
        out: Optional[np.ndarray] = None,
        mode: Optional[str] = "raise",
    ) -> PassthroughTensor:
        return self.__class__(
            self.child.choose(
                choices,
                out=out,
                mode=mode,
            )
        )

    def astype(self, np_type: np.dtype) -> PassthroughTensor:
        return self.__class__(self.child.astype(np_type))

    def __array_function__(
        self,
        func: Callable,
        types: List[Type],  # what this means =  List of Type(Type())
        args: List[Any],
        kwargs: Dict[str, Any],
    ) -> PassthroughTensor:
        # Note: this allows subclasses that don't override
        # __array_function__ to handle PassthroughTensor objects.
        if not all(issubclass(t, self.__class__) for t in types):
            return NotImplemented

        implementation = query_implementation(self.__class__, func)
        if implementation:
            return implementation(*args, **kwargs)
        return self.__class__(func(*args, **kwargs))

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        implementation = query_implementation(self.__class__, ufunc)
        if implementation:
            return implementation(*inputs, **kwargs)
        return self.__class__(ufunc(*inputs, **kwargs))

    def __repr__(self):
        return f"{self.__class__.__name__}(child={self.child})"


@implements(PassthroughTensor, np.square)
def square(x: Type[PassthroughTensor]) -> PassthroughTensor:
    return x * x



# future
from __future__ import annotations

# stdlib
from typing import Optional
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from pymbolic import var

# syft absolute
import syft as sy

# relative
from ....proto.core.adp.scalar_pb2 import PhiScalar as PhiScalar_PB
from ...common import UID
from ...common.serde.serializable import serializable
from ..entity import DataSubjectGroup
from ..entity import Entity
from ..search import ssid2obj
from .abstract.base_scalar import BaseScalar
from .intermediate_phi_scalar import IntermediatePhiScalar


@serializable()
class PhiScalar(BaseScalar, IntermediatePhiScalar):
    """
    Scalar with data from a single entity. Uses all the operations implemented in
    IntermediatePhiScalar to let the user perform operations on the data. Uses the
    BaseScalar class attributes are used to represent the data loaded in by a data owner.
    Builds on both of the above by adding SSIDs to allow the object to be referenced in
    string form and thus use the underlying polynomial libraries.
    """

    def __init__(
        self,
        min_val: Optional[float],
        value: Optional[float],
        max_val: Optional[float],
        entity: Optional[Union[Entity, DataSubjectGroup]] = None,
        id: Optional[UID] = None,
        ssid: Optional[str] = None,
    ) -> None:
        super().__init__(
            min_val=min_val, value=value, max_val=max_val, entity=entity, id=id  # type: ignore
        )
        # The scalar string identifier (SSID) - because we're using polynomial libraries
        # we need to be able to reference this object in string form. The library
        # doesn't know how to process things that aren't strings
        if ssid is None:
            ssid = "_" + self.id.no_dash + "_" + self.entity.to_string()

        self.ssid = ssid

        IntermediatePhiScalar.__init__(
            self, poly=var(self.ssid), entity=self.entity, id=self.id
        )

        ssid2obj[self.ssid] = self

    def _object2proto(self) -> PhiScalar_PB:
        kwargs = {
            "id": sy.serialize(self.id, to_proto=True),
            "entity": sy.serialize(self.entity, to_proto=True),
        }

        for field in ["max_val", "min_val", "value"]:
            if getattr(self, field) is not None:
                kwargs[field] = getattr(self, field)

        return PhiScalar_PB(**kwargs)

    @staticmethod
    def _proto2object(proto: PhiScalar_PB) -> "PhiScalar":
        return PhiScalar(
            id=sy.deserialize(proto.id),
            entity=sy.deserialize(proto.entity),
            min_val=proto.min_val if proto.HasField("min_val") else None,
            max_val=proto.max_val if proto.HasField("max_val") else None,
            value=proto.value if proto.HasField("value") else None,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return PhiScalar_PB



# stdlib
import uuid

# third party
import numpy as np

# relative
from ...passthrough import is_acceptable_simple_type  # type: ignore
from ..tensor import AutogradTensor
from .op import Op


class PowOp(Op):
    def forward(self, x: AutogradTensor, y: AutogradTensor) -> AutogradTensor:  # type: ignore
        self.x = x
        self.y = y

        requires_grad = x.requires_grad

        if is_acceptable_simple_type(y):
            return AutogradTensor(x.child**y, requires_grad=requires_grad)

        requires_grad = requires_grad or y.requires_grad
        return AutogradTensor(x.child**y.child, requires_grad=requires_grad)

    def _backward(self, grad: AutogradTensor, backprop_id: uuid.UUID) -> None:

        y_is_simple = is_acceptable_simple_type(self.y)

        if self.x.requires_grad:
            y_form = self.y

            # ignoring type b/c method hasn't been implemented yet
            self.x.add_grad(grad * y_form * (self.x ** (y_form - 1)))  # type: ignore

            if self.x.grad_fn:
                self.x.backward(backprop_id=backprop_id)

        if not y_is_simple and self.y.requires_grad:

            self.y.add_grad(np.log(self.x) * grad * self.x**self.y)

            if self.y.grad_fn:
                self.y.backward(backprop_id=backprop_id)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/adp/phi_tensor.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.adp import (
    scalar_manager_pb2 as proto_dot_core_dot_adp_dot_scalar__manager__pb2,
)
from syft.proto.core.adp import entity_pb2 as proto_dot_core_dot_adp_dot_entity__pb2
from syft.proto.core.tensor import (
    share_tensor_pb2 as proto_dot_core_dot_tensor_dot_share__tensor__pb2,
)
from syft.proto.lib.numpy import array_pb2 as proto_dot_lib_dot_numpy_dot_array__pb2
from syft.proto.lib.python import none_pb2 as proto_dot_lib_dot_python_dot_none__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1fproto/core/adp/phi_tensor.proto\x12\rsyft.core.adp\x1a\x1bproto/core/adp/entity.proto\x1a\x1bproto/lib/numpy/array.proto\x1a#proto/core/adp/scalar_manager.proto\x1a$proto/core/tensor/share_tensor.proto\x1a\x1bproto/lib/python/none.proto"\xf5\x02\n\x15SingleEntityPhiTensor\x12/\n\x06tensor\x18\x01 \x01(\x0b\x32\x1d.syft.core.tensor.ShareTensorH\x00\x12+\n\x05\x61rray\x18\x02 \x01(\x0b\x32\x1a.syft.lib.numpy.NumpyProtoH\x00\x12\'\n\x04none\x18\x03 \x01(\x0b\x32\x17.syft.lib.python.SyNoneH\x00\x12,\n\x08min_vals\x18\x04 \x01(\x0b\x32\x1a.syft.lib.numpy.NumpyProto\x12,\n\x08max_vals\x18\x05 \x01(\x0b\x32\x1a.syft.lib.numpy.NumpyProto\x12%\n\x06\x65ntity\x18\x06 \x01(\x0b\x32\x15.syft.core.adp.Entity\x12I\n\x0escalar_manager\x18\x07 \x01(\x0b\x32\x31.syft.core.adp.VirtualMachinePrivateScalarManagerB\x07\n\x05\x63hild"\xb4\x02\n\x12RowEntityPhiTensor\x12\x19\n\x11serde_concurrency\x18\x01 \x01(\r\x12\x0c\n\x04rows\x18\x02 \x03(\x0c\x12.\n\x0funique_entities\x18\x03 \x03(\x0b\x32\x15.syft.core.adp.Entity\x12Q\n\x16unique_scalar_managers\x18\x04 \x03(\x0b\x32\x31.syft.core.adp.VirtualMachinePrivateScalarManager\x12\x34\n\x10row_entity_index\x18\x05 \x01(\x0b\x32\x1a.syft.lib.numpy.NumpyProto\x12<\n\x18row_scalar_manager_index\x18\x06 \x01(\x0b\x32\x1a.syft.lib.numpy.NumpyProtob\x06proto3'
)


_SINGLEENTITYPHITENSOR = DESCRIPTOR.message_types_by_name["SingleEntityPhiTensor"]
_ROWENTITYPHITENSOR = DESCRIPTOR.message_types_by_name["RowEntityPhiTensor"]
SingleEntityPhiTensor = _reflection.GeneratedProtocolMessageType(
    "SingleEntityPhiTensor",
    (_message.Message,),
    {
        "DESCRIPTOR": _SINGLEENTITYPHITENSOR,
        "__module__": "proto.core.adp.phi_tensor_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.adp.SingleEntityPhiTensor)
    },
)
_sym_db.RegisterMessage(SingleEntityPhiTensor)

RowEntityPhiTensor = _reflection.GeneratedProtocolMessageType(
    "RowEntityPhiTensor",
    (_message.Message,),
    {
        "DESCRIPTOR": _ROWENTITYPHITENSOR,
        "__module__": "proto.core.adp.phi_tensor_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.adp.RowEntityPhiTensor)
    },
)
_sym_db.RegisterMessage(RowEntityPhiTensor)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _SINGLEENTITYPHITENSOR._serialized_start = 213
    _SINGLEENTITYPHITENSOR._serialized_end = 586
    _ROWENTITYPHITENSOR._serialized_start = 589
    _ROWENTITYPHITENSOR._serialized_end = 897
# @@protoc_insertion_point(module_scope)



# third party
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import LargeBinary

# relative
from . import Base


class PDFObject(Base):
    __tablename__ = "daa_pdf"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    binary = Column(LargeBinary(3072))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/python/collections/ordered_dict.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n/proto/lib/python/collections/ordered_dict.proto\x12\x1bsyft.lib.python.collections\x1a%proto/core/common/common_object.proto"N\n\x0bOrderedDict\x12\x0c\n\x04keys\x18\x01 \x03(\x0c\x12\x0e\n\x06values\x18\x02 \x03(\x0c\x12!\n\x02id\x18\x03 \x01(\x0b\x32\x15.syft.core.common.UIDb\x06proto3'
)


_ORDEREDDICT = DESCRIPTOR.message_types_by_name["OrderedDict"]
OrderedDict = _reflection.GeneratedProtocolMessageType(
    "OrderedDict",
    (_message.Message,),
    {
        "DESCRIPTOR": _ORDEREDDICT,
        "__module__": "proto.lib.python.collections.ordered_dict_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.python.collections.OrderedDict)
    },
)
_sym_db.RegisterMessage(OrderedDict)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _ORDEREDDICT._serialized_start = 119
    _ORDEREDDICT._serialized_end = 197
# @@protoc_insertion_point(module_scope)



# stdlib
from typing import Optional
from uuid import UUID

# third party
from numpy import ndarray

# relative
from ....common.serde.serializable import serializable
from ..tensor import AutogradTensor
from .op import Op


@serializable(recursive_serde=True)
class RepeatOp(Op):
    """Repeat operation across a dimension"""

    __attr_allowlist__ = ["x", "axis", "repeats", "input_shape", "output_shape"]

    def forward(  # type: ignore
        self, x: AutogradTensor, repeats: int, axis: Optional[int] = None
    ) -> AutogradTensor:
        self.x = x
        self.repeats = repeats
        self.axis = axis

        self.input_shape = self.x.shape

        output = x.child.repeat(repeats=repeats, axis=axis)

        self.output_shape = output.shape

        return AutogradTensor(output, requires_grad=x.requires_grad)

    def _backward(self, grad: ndarray, backprop_id: UUID) -> None:

        if self.x.requires_grad:

            axis = self.axis
            if axis is None:
                axis = len(self.input_shape) - 1

            intermediate_shape = list(self.input_shape)
            intermediate_shape.insert(axis + 1, -1)

            if self.x.shape[self.axis] == 1:  # type: ignore
                grad = grad.sum(axis=axis)
            else:
                grad = grad.reshape(*intermediate_shape)
                grad = grad.sum(axis=axis + 1)

            self.x.add_grad(grad)

            if self.x.grad_fn:
                self.x.backward(backprop_id=backprop_id)



# stdlib
import math
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

# third party
import boto3
from nacl.signing import VerifyKey
from typing_extensions import final

# relative
from ......grid import GridURL
from .....common.serde.serializable import serializable
from ....domain.domain_interface import DomainInterface
from ....domain.registry import DomainMessageRegistry
from ...permissions.permissions import BasePermission
from ...permissions.user_permissions import NoRestriction
from ...util import get_s3_client
from ..generic_payload.syft_message import NewSyftMessage as SyftMessage
from ..generic_payload.syft_message import ReplyPayload
from ..generic_payload.syft_message import RequestPayload


@serializable(recursive_serde=True)
@final
class UploadDataMessage(SyftMessage, DomainMessageRegistry):

    # Pydantic Inner class to define expected request payload fields.
    class Request(RequestPayload):
        """Payload fields and types used during a User Creation Request."""

        filename: str
        file_size: int
        chunk_size: int
        content_type: Optional[str]

    # Pydantic Inner class to define expected reply payload fields.
    class Reply(ReplyPayload):
        """Payload fields and types used during a User Creation Response."""

        upload_id: str
        parts: List[Dict[Any, Any]]

    request_payload_type = (
        Request  # Converts generic syft dict into a CreateUserMessage.Request object.
    )
    reply_payload_type = (
        Reply  # Creates a proper Reply payload message structure as a response.
    )

    def run(  # type: ignore
        self, node: DomainInterface, verify_key: Optional[VerifyKey] = None
    ) -> ReplyPayload:  # type: ignore
        # TODO : Move to permissions
        # if not node.users.can_upload_data(verify_key=verify_key):
        #    return {"message": "You're not authorized to do this."}

        key = f"{self.payload.filename}"
        s3_client = get_s3_client(settings=node.settings)
        result = s3_client.create_multipart_upload(Bucket=node.id.no_dash, Key=key)
        total_parts = math.ceil(self.payload.file_size / self.payload.chunk_size)
        upload_id = result["UploadId"]

        parts = list()
        for part_no in range(1, total_parts + 1):
            # Creating presigned urls
            signed_url = s3_client.generate_presigned_url(
                ClientMethod="upload_part",
                Params={
                    "Bucket": node.id.no_dash,
                    "Key": key,
                    "UploadId": upload_id,
                    "PartNumber": part_no,
                },
                ExpiresIn=node.settings.S3_PRESIGNED_TIMEOUT_SECS,
            )
            # parse as a URL
            grid_url = GridURL.from_url(url=signed_url)
            # add /blob to path
            grid_url.path = f"/blob{grid_url.path}"
            # only return the path and let the client use its existing public url
            parts.append({"part_no": part_no, "url": grid_url.url_path})

        return UploadDataMessage.Reply(upload_id=upload_id, parts=parts)

    def get_permissions(self) -> List[Type[BasePermission]]:
        """Returns the list of permission classes."""
        return [NoRestriction]


@serializable(recursive_serde=True)
@final
class UploadDataCompleteMessage(SyftMessage, DomainMessageRegistry):

    # Pydantic Inner class to define expected request payload fields.
    class Request(RequestPayload):
        """Payload fields and types used during a User Creation Request."""

        filename: str
        upload_id: str
        parts: List[Dict[Any, Any]]

    # Pydantic Inner class to define expected reply payload fields.
    class Reply(ReplyPayload):
        """Payload fields and types used during a User Creation Response."""

        message: str = "Upload Complete!"

    request_payload_type = (
        Request  # Converts generic syft dict into a CreateUserMessage.Request object.
    )
    reply_payload_type = (
        Reply  # Creates a proper Reply payload message structure as a response.
    )

    def run(  # type: ignore
        self, node: DomainInterface, verify_key: Optional[VerifyKey] = None
    ) -> ReplyPayload:  # type: ignore

        # TODO: Move to permissions
        # user_role = node.roles.first(**{"id": current_user.role})
        # if not user_role.can_upload_data:
        #    return {"message": "You're not authorized to do this."}

        key = f"{self.payload.filename}"
        client: boto3.client.S3 = get_s3_client(settings=node.settings)
        _ = client.complete_multipart_upload(
            Bucket=node.id.no_dash,
            Key=key,
            MultipartUpload={"Parts": self.payload.parts},
            UploadId=self.payload.upload_id,
        )
        return UploadDataCompleteMessage.Reply()

    def get_permissions(self) -> List[Type[BasePermission]]:
        """Returns the list of permission classes."""
        return [NoRestriction]



# stdlib
from typing import Any
from typing import Optional
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from pymbolic import var

# syft absolute
import syft as sy

# relative
from ....proto.core.adp.scalar_pb2 import GammaScalar as GammaScalar_PB
from ...common import UID
from ...common.serde.serializable import serializable
from ..entity import DataSubjectGroup
from ..entity import Entity
from ..search import ssid2obj
from .abstract.base_scalar import BaseScalar
from .intermediate_gamma_scalar import IntermediateGammaScalar


@serializable()
class GammaScalar(BaseScalar, IntermediateGammaScalar):
    """
    A scalar over data from multiple entities. Uses all the operators from IntermediateGammaScalar.
    Uses BaseScalar to represent the node at which point data from multiple entities was combined.
    Finally, adds SSIDs to allow the underlying polynomial libraries to work and run.
    """

    def __init__(
        self,
        min_val: float,
        value: float,
        max_val: float,
        prime: int,
        entity: Optional[Union[Entity, DataSubjectGroup]] = None,
        id: Optional[UID] = None,
        ssid: Optional[str] = None,
    ) -> None:
        super().__init__(
            min_val=min_val, value=value, max_val=max_val, entity=entity, id=id
        )

        self.prime = prime

        # The scalar string identifier (SSID) - because we're using polynomial libraries
        # we need to be able to reference this object in string form. The library
        # doesn't know how to process things that aren't strings
        if ssid is None:
            ssid = "_" + self.id.no_dash + "_" + self.entity.to_string()
        self.ssid = ssid

        IntermediateGammaScalar.__init__(
            self, poly=var(self.ssid), min_val=min_val, max_val=max_val, id=self.id
        )

        ssid2obj[self.ssid] = self

    def _object2proto(self) -> GammaScalar_PB:
        kwargs = {
            "id": sy.serialize(self.id, to_proto=True),
            "prime": self.prime,
        }
        entity_data = sy.serialize(self.entity, to_proto=True)
        if isinstance(self.entity, Entity):
            kwargs["entity"] = entity_data
        elif isinstance(self.entity, DataSubjectGroup):
            kwargs["dsg"] = entity_data
        else:
            raise ValueError(
                f"Invalid type: {type(self.entity)} in GammaScalar Entity field."
            )

        for field in ["max_val", "min_val", "value"]:
            if getattr(self, field) is not None:
                kwargs[field] = getattr(self, field)

        return GammaScalar_PB(**kwargs)

    @staticmethod
    def _proto2object(proto: GammaScalar_PB) -> "GammaScalar":
        if proto.HasField("entity"):
            entity_data = sy.deserialize(proto.entity)
        elif proto.HasField("dsg"):
            entity_data = sy.deserialize(proto.dsg)
        return GammaScalar(
            id=sy.deserialize(proto.id, from_proto=True),
            min_val=proto.min_val if proto.HasField("min_val") else None,
            max_val=proto.max_val if proto.HasField("max_val") else None,
            value=proto.value if proto.HasField("value") else None,
            entity=entity_data,
            prime=proto.prime,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return GammaScalar_PB

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, GammaScalar):
            return (
                self.min_val == other.min_val
                and self.max_val == other.max_val
                and self.value == other.value
                and self.entity == other.entity
                and self.prime == other.prime
            )
        return self == other

    def __hash__(self) -> int:
        return (
            hash(self.min_val)
            + hash(self.max_val)
            + hash(self.prime)
            + hash(self.entity)
            + hash(self.value)
        )



# # future
# from __future__ import annotations
#
# # stdlib
# from typing import List as TypeList
# from typing import Union
#
# # third party
# import names
# import numpy as np
#
# # relative
# from ..adp.entity import Entity
# from ..adp.publish import publish
# from ..adp.scalar import PhiScalar
# from .autodp.single_entity_phi import SingleEntityPhiTensor
# from .autodp.row_entity_phi import RowEntityPhiTensor
#
#
#
# def make_entities(n: int = 100) -> TypeList[Entity]:
#     ents: TypeList[Entity] = list()
#     for i in range(n):
#         ents.append(Entity(name=names.get_full_name().replace(" ", "_")))
#     return ents
#
#
# def private(
#     input_data,
#     min_val,
#     max_val,
#     entities=None,
#     one_entity_per_row=True,
#     is_discrete=False,
# ) -> Union[SingleEntityPhiTensor, RowEntityPhiTensor, ]:
#     self = input_data
#
#     if one_entity_per_row and entities is None:
#         entities = make_entities(n=len(input_data))
#
#     if entities is None:
#         flat_data = self.flatten()
#         entities = make_entities(n=len(flat_data))
#
#         scalars = list()
#         for i in flat_data:
#             value = max(min(float(i), max_val), min_val)
#             s = PhiScalar(
#                 value=value,
#                 min_val=min_val,
#                 max_val=max_val,
#                 entity=entities[len(scalars)],
#                 #                 is_discrete=is_discrete
#             )
#             scalars.append(s)
#
#         return to_values(np.array(scalars)).reshape(input_data.shape)
#
#     elif isinstance(entities, list):
#         if len(entities) == len(self):
#             output_rows = list()
#             for row_i, row in enumerate(self):
#                 row_of_entries = list()
#                 for item in row.flatten():
#                     s = PhiScalar(
#                         value=item,
#                         min_val=min_val,
#                         max_val=max_val,
#                         entity=entities[row_i],
#                         #                         is_discrete=is_discrete
#                     )
#                     row_of_entries.append(s)
#                 output_rows.append(np.array(row_of_entries).reshape(row.shape))
#             return to_values(np.array(output_rows)).reshape(self.shape)
#         else:
#             print(len(entities))
#             print(len(self))
#             raise Exception("len(entities) must equal len(self)")
#
#
# class GradLedger:
#     def __init__(self):
#         self.grad_parents = list()
#         self.total_symbols = set()
#
#     def add(self, dict_of_gradient_parents, accumulate_grads=False):
#
#         self.grad_parents.append(dict_of_gradient_parents)
#
#         # if accumulate grads is false, raise exception if any of the results
#         # set gradients for the same variables (which should have been accumulated)
#
#         for symbol_id in dict_of_gradient_parents.keys():
#             if not accumulate_grads:
#                 if symbol_id in self.total_symbols:
#                     exception_msg = (  # nosec
#                         "You had multiple gradients attempting to update the same"
#                         + "scalar but accumulate_grads was set to False. This means that"
#                         + "some grads were overwritten by new ones and these gradients"
#                         + "would be false. Please re-run the computation with"
#                         + "accumulate_grads set to True and don't forget to zero grads"
#                         + "out each time you're done."
#                     )
#                     raise Exception(exception_msg)
#             self.total_symbols.add(symbol_id)
#
#     def zero_grads(self):
#         for x in self.total_symbols:
#             x._grad = None
#
#
# class ScalarTensor(np.ndarray):
#     def __new__(
#         cls,
#         input_array,
#         min_val=None,
#         max_val=None,
#         entities=None,
#         info=None,
#         is_discrete=False,
#     ):
#
#         is_private = False
#
#         if min_val is not None and max_val is not None:
#             input_array = private(
#                 input_array,
#                 min_val=min_val,
#                 max_val=max_val,
#                 entities=entities,
#                 is_discrete=is_discrete,
#             )
#             is_private = True
#         else:
#             input_array = to_values(input_array)
#
#         obj = np.asarray(input_array).view(cls)
#         obj.info = info
#         obj.is_private = is_private
#
#         return obj
#
#     def __array_finalize__(self, obj):
#         if obj is None:
#             return
#         self.info = getattr(obj, "info", None)
#         self.is_private = getattr(obj, "is_private", None)
#
#     def __array_wrap__(self, out_arr, context=None):
#
#         output = out_arr.view(ScalarTensor)
#
#         is_private = False
#         if context is not None:
#             for arg in context[1]:
#                 if hasattr(arg, "is_private") and arg.is_private:
#                     is_private = True
#
#         output.is_private = is_private
#
#         return output
#
#     def backward(self, accumulate_grads=False) -> GradLedger:
#         ledger = GradLedger()
#         for entry in self.flatten():
#             ledger.add(
#                 grad(entry, accumulate=accumulate_grads),
#                 accumulate_grads=accumulate_grads,
#             )
#         return ledger
#
#     @property
#     def grad(self):
#         grads = list()
#         for val in self.flatten().tolist():
#             grads.append(val._grad)
#         return ScalarTensor(grads).reshape(self.shape)
#
#     def slow_publish(self, **kwargs):
#         grads = list()
#         for val in self.flatten().tolist():
#             grads.append(val.value.publish(**kwargs))
#         return np.array(grads).reshape(self.shape)
#
#     def publish(self, **kwargs):
#         grads = list()
#         for val in self.flatten().tolist():
#             grads.append(val.value)
#         grads = publish(scalars=grads, **kwargs)
#         return np.array(grads).reshape(self.shape)
#
#     @property
#     def value(self):
#         values = list()
#         for val in self.flatten().tolist():
#             if hasattr(val.value, "value"):
#                 values.append(val.value.value)
#             else:
#                 values.append(val.value)
#         return np.array(values).reshape(self.shape)
#
#     def private(self, min_val, max_val, entities=None, is_discrete=False):
#         if self.is_private:
#             raise Exception("Cannot call .private() on tensor which is already private")
#
#         return ScalarTensor(
#             self.value,
#             min_val=min_val,
#             max_val=max_val,
#             entities=entities,
#             is_discrete=is_discrete,
#         )



# third party
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import JSON
from sqlalchemy import String

# relative
from . import Base


class ObjectMetadata(Base):
    __tablename__ = "obj_metadata"

    __mapper_args__ = {"confirm_deleted_rows": False}

    id = Column(Integer(), primary_key=True, autoincrement=True)
    obj = Column(String(256))
    tags = Column(JSON())
    description = Column(String())
    name = Column(String())
    read_permissions = Column(JSON())
    search_permissions = Column(JSON())
    write_permissions = Column(JSON())



# CLEANUP NOTES:
# - remove unused comments
# - add documentation for each method
# - add comments inline explaining each piece
# - add a unit test for each method (at least)

# stdlib
from functools import lru_cache
from typing import Dict
from typing import List
from typing import Optional

# third party
from autodp import dp_bank
from autodp import fdp_bank
from autodp.autodp_core import Mechanism
from nacl.signing import VerifyKey
import numpy as np

# relative
from ..common.serde.serializable import serializable


# methods serialize/deserialize np.int64 number
# syft.serde seems to not support np.int64 serialization/deserialization
def numpy64tolist(value: np.int64) -> List:
    list_version = value.tolist()
    return list_version


def listtonumpy64(value: List) -> np.int64:
    return np.int64(value)


# returns the privacy budget spent by each entity
@lru_cache(maxsize=None)
def _individual_RDP_gaussian(
    sigma: float, value: float, L: float, alpha: float
) -> float:
    return (alpha * (L**2) * (value**2)) / (2 * (sigma**2))


def individual_RDP_gaussian(params: Dict, alpha: float) -> np.float64:
    """
    :param params:
        'sigma' --- is the normalized noise level: std divided by global L2 sensitivity
        'value' --- is the output of query on a data point
        'L' --- is the Lipschitz constant of query with respect to the output of query on a data point
    :param alpha: The order of the Renyi Divergence
    :return: Evaluation of the RDP's epsilon
    """
    sigma = params["sigma"]
    value = params["value"]
    L = params["L"]
    if sigma <= 0:
        raise Exception("Sigma should be above 0")
    if alpha < 0:
        raise Exception("Sigma should not be below 0")

    return _individual_RDP_gaussian(sigma=sigma, alpha=alpha, value=value, L=L)


# Example of a specific mechanism that inherits the Mechanism class
@serializable(recursive_serde=True)
class iDPGaussianMechanism(Mechanism):
    __attr_allowlist__ = [
        "name",
        "params",
        "entity_name",
        "fdp",
        "eps_pureDP",
        "delta0",
        "RDP_off",
        "approxDP_off",
        "fdp_off",
        "use_basic_rdp_to_approx_dp_conversion",
        "use_fdp_based_rdp_to_approx_dp_conversion",
        "user_key",
    ]

    # delta0 is a numpy.int64 number (not supported by syft.serde)
    __serde_overrides__ = {
        "delta0": [numpy64tolist, listtonumpy64],
    }

    def __init__(
        self,
        sigma: float,
        squared_l2_norm: float,
        squared_l2_norm_upper_bound: float,
        L: float,
        entity_name: str,
        name: str = "Gaussian",
        RDP_off: bool = False,
        approxDP_off: bool = False,
        fdp_off: bool = True,
        use_basic_rdp_to_approx_dp_conversion: bool = False,
        use_fdp_based_rdp_to_approx_dp_conversion: bool = False,
        user_key: Optional[VerifyKey] = None,
    ):

        # the sigma parameter is the std of the noise divide by the l2 sensitivity
        Mechanism.__init__(self)

        self.user_key = user_key

        self.name = name  # When composing
        self.params = {
            "sigma": float(sigma),
            "private_value": float(squared_l2_norm),
            "public_value": float(squared_l2_norm_upper_bound),
            "L": float(L),
        }  # This will be useful for the Calibrator

        self.entity_name = entity_name
        # TODO: should a generic unspecified mechanism have a name and a param dictionary?

        self.delta0 = 0
        if not RDP_off:
            # Tudor: i'll fix these
            new_rdp = lambda x: individual_RDP_gaussian(self.params, x)  # noqa: E731
            if use_fdp_based_rdp_to_approx_dp_conversion:
                # This setting is slightly more complex, which involves converting RDP to fDP,
                # then to eps-delta-DP via the duality
                self.propagate_updates(new_rdp, "RDP", fDP_based_conversion=True)
            elif use_basic_rdp_to_approx_dp_conversion:
                self.propagate_updates(new_rdp, "RDP", BBGHS_conversion=False)
            else:
                # This is the default setting with fast computation of RDP to approx-DP
                self.propagate_updates(new_rdp, "RDP")

        if not approxDP_off:  # Direct implementation of approxDP
            new_approxdp = lambda x: dp_bank.get_eps_ana_gaussian(  # noqa: E731
                sigma, x
            )
            self.propagate_updates(new_approxdp, "approxDP_func")

        if not fdp_off:  # Direct implementation of fDP
            # Tudor: i'll fix these
            fun1 = lambda x: fdp_bank.log_one_minus_fdp_gaussian(  # noqa: E731
                {"sigma": sigma}, x
            )
            fun2 = lambda x: fdp_bank.log_neg_fdp_grad_gaussian(  # noqa: E731
                {"sigma": sigma}, x
            )
            self.propagate_updates([fun1, fun2], "fDP_and_grad_log")
            # overwrite the fdp computation with the direct computation
            self.fdp = lambda x: fdp_bank.fDP_gaussian(
                {"sigma": sigma}, x
            )  # noqa: E731

        # the fDP of gaussian mechanism is equivalent to analytical calibration of approxdp,
        # so it should have been automatically handled numerically above

        # Discussion:  Sometimes delta as a function of eps has a closed-form solution
        # while eps as a function of delta does not
        # Shall we represent delta as a function of eps instead?
